"""
GUI module for training application.

This module provides the `TrainGUI` class that handles the graphical user interface (GUI) for managing
training workflows in Supervisely.
"""

import json
import os
from os import environ, getenv
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import supervisely.io.env as sly_env
import supervisely.io.fs as sly_fs
import supervisely.io.json as sly_json
from supervisely import Api, ProjectMeta, logger
from supervisely._utils import is_production
from supervisely.app.widgets import Button, Card, Stepper, Widget
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.experiments import ExperimentInfo
from supervisely.nn.task_type import TaskType
from supervisely.nn.training.gui.classes_selector import ClassesSelector
from supervisely.nn.training.gui.hyperparameters_selector import HyperparametersSelector
from supervisely.nn.training.gui.input_selector import InputSelector
from supervisely.nn.training.gui.model_selector import ModelSelector
from supervisely.nn.training.gui.tags_selector import TagsSelector
from supervisely.nn.training.gui.train_val_splits_selector import TrainValSplitsSelector
from supervisely.nn.training.gui.training_artifacts import TrainingArtifacts
from supervisely.nn.training.gui.training_logs import TrainingLogs
from supervisely.nn.training.gui.training_process import TrainingProcess
from supervisely.nn.training.gui.utils import set_stepper_step, wrap_button_click
from supervisely.nn.utils import ModelSource, RuntimeType


class StepFlow:
    """
    Manages the flow of steps in the GUI, including wrappers and button handlers.

    Allows flexible configuration of dependencies between steps and automatically
    sets up proper handlers based on layout from app_options.
    """

    def __init__(self, stepper: Stepper, app_options: Dict[str, Any]):
        """
        Initializes the step manager.

        :param stepper: :class:`~supervisely.app.widgets.stepper.stepper.Stepper` object for step navigation
        :param app_options: Application options
        """
        self.stepper = stepper
        self.app_options = app_options
        self.collapsable = app_options.get("collapsable", False)
        self.steps = {}  # Step configuration
        self.step_sequence = []  # Step sequence

    def register_step(
        self,
        name: str,
        card: Card,
        button: Optional[Button] = None,
        widgets_to_disable: Optional[List[Widget]] = None,
        validation_text: Optional[Widget] = None,
        validation_func: Optional[Callable] = None,
        position: Optional[int] = None,
    ) -> "StepFlow":
        """
        Registers a step in the GUI.

        :param name: Unique step name
        :param card: Step card widget
        :param button: Button for proceeding to the next step (optional)
        :param widgets_to_disable: Widgets to disable during validation (optional)
        :param validation_text: Widget for displaying validation text (optional)
        :param validation_func: Validation function (optional)
        :param position: Step position in the sequence (starting from 0)
        :returns: Current StepFlow object for method chaining
        """
        self.steps[name] = {
            "card": card,
            "button": button,
            "widgets_to_disable": widgets_to_disable or [],
            "validation_text": validation_text,
            "validation_func": validation_func,
            "position": position,
            "next_steps": [],
            "on_select_click": [],
            "on_reselect_click": [],
            "wrapper": None,
            "has_button": button is not None,
        }

        if position is not None:
            while len(self.step_sequence) <= position:
                self.step_sequence.append(None)
            self.step_sequence[position] = name

        return self

    def set_next_steps(self, step_name: str, next_steps: List[str]) -> "StepFlow":
        """
        Sets the list of next steps for the given step.

        :param step_name: Current step name
        :param next_steps: List of names of the next steps
        :returns: Current StepFlow object for method chaining
        """
        if step_name in self.steps:
            self.steps[step_name]["next_steps"] = next_steps
        return self

    def add_on_select_actions(
        self, step_name: str, actions: List[Callable], is_reselect: bool = False
    ) -> "StepFlow":
        """
        Adds actions to be executed when a step is selected/reselected.

        :param step_name: Step name
        :param actions: List of functions to execute
        :param is_reselect: True if these are actions for reselection, otherwise False
        :returns: Current StepFlow object for method chaining
        """
        if step_name in self.steps:
            key = "on_reselect_click" if is_reselect else "on_select_click"
            self.steps[step_name][key].extend(actions)
        return self

    def build_wrappers(self) -> Dict[str, Callable]:
        """
        Creates wrappers for all steps based on established dependencies.

        :returns: Dictionary with created wrappers by step name
        """
        valid_sequence = [s for s in self.step_sequence if s is not None and s in self.steps]

        for step_name in reversed(valid_sequence):
            step = self.steps[step_name]

            cards_to_unlock = []
            for next_step_name in step["next_steps"]:
                if next_step_name in self.steps:
                    cards_to_unlock.append(self.steps[next_step_name]["card"])

            callback = None
            if step["next_steps"] and step["has_button"]:
                for next_step_name in step["next_steps"]:
                    if (
                        next_step_name in self.steps
                        and self.steps[next_step_name].get("wrapper")
                        and self.steps[next_step_name]["has_button"]
                    ):
                        callback = self.steps[next_step_name]["wrapper"]
                        break

            if step["has_button"]:
                wrapper = wrap_button_click(
                    button=step["button"],
                    cards_to_unlock=cards_to_unlock,
                    widgets_to_disable=step["widgets_to_disable"],
                    callback=callback,
                    validation_text=step["validation_text"],
                    validation_func=step["validation_func"],
                    on_select_click=step["on_select_click"],
                    on_reselect_click=step["on_reselect_click"],
                    collapse_card=(step["card"], self.collapsable),
                )

                step["wrapper"] = wrapper

        return {
            name: self.steps[name]["wrapper"]
            for name in self.steps
            if self.steps[name].get("wrapper") and self.steps[name]["has_button"]
        }

    def setup_button_handlers(self) -> None:
        """
        Sets up handlers for buttons of all steps.
        """
        positions = {}
        pos = 1

        for i, step_name in enumerate(self.step_sequence):
            if step_name is not None and step_name in self.steps:
                positions[step_name] = pos
                pos += 1

        for step_name, step in self.steps.items():
            if step_name in positions and step.get("wrapper") and step["has_button"]:

                button = step["button"]
                wrapper = step["wrapper"]
                position = positions[step_name]
                next_position = position + 1

                def create_handler(btn, cb, next_pos):
                    def handler():
                        cb()
                        set_stepper_step(self.stepper, btn, next_pos=next_pos)

                    return handler

                button.click(create_handler(button, wrapper, next_position))

    def build(self) -> Dict[str, Callable]:
        """
        Performs the complete setup of the step system.

        :returns: Dictionary with created wrappers by step name
        """
        wrappers = self.build_wrappers()
        self.setup_button_handlers()
        return wrappers


class TrainGUI:
    """GUI for training workflows: project, split, model, hyperparameters, and training process."""

    def __init__(
        self,
        framework_name: str,
        models: list,
        hyperparameters: dict,
        app_options: dict = None,
    ):
        """:param framework_name: Name of the ML framework being used.
        :type framework_name: str
        :param models: List of available models.
        :type models: list
        :param hyperparameters: Hyperparameters for training.
        :type hyperparameters: dict
        :param app_options: Application options for customization.
        :type app_options: dict, optional
        """
        self._api = Api.from_env()
        if is_production():
            self.task_id = sly_env.task_id()
        else:
            self.task_id = sly_env.task_id(raise_not_found=False)
            if self.task_id is None:
                self.task_id = -1

        self.framework_name = framework_name
        self.models = models
        self.hyperparameters = hyperparameters
        self.app_options = app_options
        self.collapsable = self.app_options.get("collapsable", False)
        self.need_convert_shapes = False
        self._start_training = False

        self.team_id = sly_env.team_id(raise_not_found=False)
        self.workspace_id = sly_env.workspace_id(raise_not_found=False)
        self.project_id = sly_env.project_id()
        self.project_info = self._api.project.get_info_by_id(self.project_id)
        if self.project_info is None:
            raise ValueError(f"Project with ID: '{self.project_id}' does not exist or was archived")

        self.project_meta = ProjectMeta.from_json(self._api.project.get_meta(self.project_id))

        if self.workspace_id is None:
            self.workspace_id = self.project_info.workspace_id
            environ["WORKSPACE_ID"] = str(self.workspace_id)
        if self.team_id is None:
            self.team_id = self.project_info.team_id
            environ["TEAM_ID"] = str(self.team_id)

        # ---------- Parse selector options ----------
        self._classes_selector_opts = self.app_options.get("classes_selector", {})
        self._tags_selector_opts = self.app_options.get("tags_selector", {})
        self._train_val_splits_opts = self.app_options.get("train_val_splits_selector", {})
        self._model_selector_opts = self.app_options.get("model_selector", {})

        self.show_classes_selector = self._classes_selector_opts.get("enabled", True)
        self.show_tags_selector = self._tags_selector_opts.get("enabled", False)
        self.show_train_val_splits_selector = self._train_val_splits_opts.get("enabled", True)
        self.show_model_selector = self._model_selector_opts.get("enabled", True)

        # Ensure train_val_splits_methods compatibility
        self._train_val_methods = self._train_val_splits_opts.get("methods", [])
        # --------------------------------------------------------- #

        # ------------------------------------------------- #
        self.steps = []

        # 1. Project selection
        self.input_selector = InputSelector(self.project_info, self.app_options)
        self.steps.append(self.input_selector.card)

        # 2. Model selector
        self.model_selector = ModelSelector(
            self._api, self.framework_name, self.models, self.app_options
        )
        if self.show_model_selector:
            self.steps.append(self.model_selector.card)

        # 3. Classes selector
        self.classes_selector = None
        if self.show_classes_selector:
            self.classes_selector = ClassesSelector(
                self.project_id, [], self.model_selector, self.app_options
            )
            self.steps.append(self.classes_selector.card)

        # 4. Tags selector
        self.tags_selector = None
        if self.show_tags_selector:
            self.tags_selector = TagsSelector(self.project_id, [], self.app_options)
            self.steps.append(self.tags_selector.card)

        # 5. Train/Val splits selector
        self.train_val_splits_selector = None
        if self.show_train_val_splits_selector:
            self.train_val_splits_selector = TrainValSplitsSelector(
                self._api, self.project_id, self.app_options
            )
            self.steps.append(self.train_val_splits_selector.card)

        # 6. Training parameters (yaml)
        self.hyperparameters_selector = HyperparametersSelector(
            self.hyperparameters, self.app_options
        )
        self.steps.append(self.hyperparameters_selector.card)

        # 7. Start Training
        self.training_process = TrainingProcess(self.app_options)
        self.steps.append(self.training_process.card)

        # 8. Training logs
        self.training_logs = TrainingLogs(self.app_options)
        self.steps.append(self.training_logs.card)

        # 9. Training Artifacts
        self.training_artifacts = TrainingArtifacts(self._api, self.app_options)
        self.steps.append(self.training_artifacts.card)

        # Stepper layout
        self.stepper = Stepper(widgets=self.steps)
        # ------------------------------------------------- #

        # Button utils
        def disable_hyperparams_editor():
            if self.hyperparameters_selector.editor.readonly:
                self.hyperparameters_selector.editor.readonly = False
            else:
                self.hyperparameters_selector.editor.readonly = True

        def set_experiment_name():
            model_name = self.model_selector.get_model_name()
            if model_name is None:
                experiment_name = "Enter experiment name"
            else:
                if self.task_id == -1:
                    experiment_name = f"debug {self.project_info.name} {model_name}"
                else:
                    experiment_name = f"{self.task_id} {self.project_info.name} {model_name}"

            if experiment_name == self.training_process.get_experiment_name():
                return
            self.training_process.set_experiment_name(experiment_name)

        def need_convert_class_shapes() -> bool:
            if (
                self.hyperparameters_selector.run_model_benchmark_checkbox is None
                or not self.hyperparameters_selector.run_model_benchmark_checkbox.is_checked()
            ):
                self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()
                self.need_convert_shapes = False
                return False

            task_type = self.model_selector.get_selected_task_type()

            if self.classes_selector is not None:
                wrong_shapes = set(self.classes_selector.get_wrong_shape_classes(task_type))

                # Exclude classes with no annotations to avoid unnecessary conversion
                data = self.classes_selector.classes_table._table_data
                empty_classes = {
                    r[0]["data"] for r in data if r[2]["data"] == 0 and r[3]["data"] == 0
                }
                need_conversion = bool(wrong_shapes - empty_classes)
            else:
                # Classes selector disabled – check entire project meta
                if task_type == TaskType.OBJECT_DETECTION:
                    need_conversion = any(
                        obj_cls.geometry_type != Rectangle
                        for obj_cls in self.project_meta.obj_classes
                    )
                elif task_type in [TaskType.INSTANCE_SEGMENTATION, TaskType.SEMANTIC_SEGMENTATION]:
                    need_conversion = any(
                        obj_cls.geometry_type == Polygon
                        for obj_cls in self.project_meta.obj_classes
                    )
                else:
                    need_conversion = False

            if need_conversion:
                self.hyperparameters_selector.model_benchmark_auto_convert_warning.show()
            else:
                self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()

            self.need_convert_shapes = need_conversion
            return need_conversion

        # ------------------------------------------------- #

        self.step_flow = StepFlow(self.stepper, self.app_options)
        position = 0

        # 1. Input selector
        self.step_flow.register_step(
            "input_selector",
            self.input_selector.card,
            self.input_selector.button,
            self.input_selector.widgets_to_disable,
            self.input_selector.validator_text,
            self.input_selector.validate_step,
            position=position,
        )
        position += 1

        # 2. Model selector
        if self.show_model_selector:
            self.step_flow.register_step(
                "model_selector",
                self.model_selector.card,
                self.model_selector.button,
                self.model_selector.widgets_to_disable,
                self.model_selector.validator_text,
                self.model_selector.validate_step,
                position=position,
            ).add_on_select_actions("model_selector", [set_experiment_name])
            position += 1

        # 3. Classes selector
        if self.show_classes_selector and self.classes_selector is not None:
            self.step_flow.register_step(
                "classes_selector",
                self.classes_selector.card,
                self.classes_selector.button,
                self.classes_selector.widgets_to_disable,
                self.classes_selector.validator_text,
                self.classes_selector.validate_step,
                position=position,
            ).add_on_select_actions("classes_selector", [need_convert_class_shapes])
            position += 1

        # 4. Tags selector
        if self.show_tags_selector and self.tags_selector is not None:
            self.step_flow.register_step(
                "tags_selector",
                self.tags_selector.card,
                self.tags_selector.button,
                self.tags_selector.widgets_to_disable,
                self.tags_selector.validator_text,
                self.tags_selector.validate_step,
                position=position,
            )
            position += 1

        # 5. Train/Val splits selector
        if self.show_train_val_splits_selector and self.train_val_splits_selector is not None:
            self.step_flow.register_step(
                "train_val_splits",
                self.train_val_splits_selector.card,
                self.train_val_splits_selector.button,
                self.train_val_splits_selector.widgets_to_disable,
                self.train_val_splits_selector.validator_text,
                self.train_val_splits_selector.validate_step,
                position=position,
            )
            position += 1

        # 6. Hyperparameters selector
        self.step_flow.register_step(
            "hyperparameters_selector",
            self.hyperparameters_selector.card,
            self.hyperparameters_selector.button,
            self.hyperparameters_selector.widgets_to_disable,
            self.hyperparameters_selector.validator_text,
            self.hyperparameters_selector.validate_step,
            position=position,
        ).add_on_select_actions("hyperparameters_selector", [disable_hyperparams_editor])
        self.step_flow.add_on_select_actions(
            "hyperparameters_selector", [disable_hyperparams_editor], is_reselect=True
        )
        position += 1

        # 7. Training process
        self.step_flow.register_step(
            "training_process",
            self.training_process.card,
            None,
            self.training_process.widgets_to_disable,
            self.training_process.validator_text,
            self.training_process.validate_step,
            position=position,
        )
        position += 1

        # 8. Training logs
        self.step_flow.register_step(
            "training_logs",
            self.training_logs.card,
            None,
            self.training_logs.widgets_to_disable,
            self.training_logs.validator_text,
            self.training_logs.validate_step,
            position=position,
        )
        position += 1

        # 9. Training artifacts
        self.step_flow.register_step(
            "training_artifacts",
            self.training_artifacts.card,
            None,
            self.training_artifacts.widgets_to_disable,
            self.training_artifacts.validator_text,
            self.training_artifacts.validate_step,
            position=position,
        )

        # Set dependencies between steps
        has_model_selector = self.show_model_selector and self.model_selector is not None
        has_classes_selector = self.show_classes_selector and self.classes_selector is not None
        has_tags_selector = self.show_tags_selector and self.tags_selector is not None
        has_train_val_splits = (
            self.show_train_val_splits_selector and self.train_val_splits_selector is not None
        )

        # Set step dependency chain
        prev_step = "input_selector"
        if has_model_selector:
            self.step_flow.set_next_steps(prev_step, ["model_selector"])
            prev_step = "model_selector"
        if has_classes_selector:
            self.step_flow.set_next_steps(prev_step, ["classes_selector"])
            prev_step = "classes_selector"
        if has_tags_selector:
            self.step_flow.set_next_steps(prev_step, ["tags_selector"])
            prev_step = "tags_selector"
        if has_train_val_splits:
            self.step_flow.set_next_steps(prev_step, ["train_val_splits"])
            prev_step = "train_val_splits"

        # 6. Hyperparameters selector -> 7. Training process
        self.step_flow.set_next_steps(prev_step, ["hyperparameters_selector"])
        self.step_flow.set_next_steps("hyperparameters_selector", ["training_process"])

        # 7. Training process -> 8. Training logs
        self.step_flow.set_next_steps("training_process", ["training_logs"])

        # 8. Training logs -> 9. Training artifacts
        self.step_flow.set_next_steps("training_logs", ["training_artifacts"])
        # ------------------------------------------------- #

        # Create all wrappers and set button handlers
        wrappers = self.step_flow.build()

        self.input_selector_cb = wrappers.get("input_selector")
        self.train_val_splits_selector_cb = wrappers.get("train_val_splits")
        self.classes_selector_cb = wrappers.get("classes_selector")
        self.tags_selector_cb = wrappers.get("tags_selector")
        self.model_selector_cb = wrappers.get("model_selector")
        self.hyperparameters_selector_cb = wrappers.get("hyperparameters_selector")
        self.training_process_cb = wrappers.get("training_process")
        self.training_logs_cb = wrappers.get("training_logs")
        self.training_artifacts_cb = wrappers.get("training_artifacts")
        # ------------------------------------------------- #

        # Other handlers
        if self.app_options.get("show_logs_in_gui", False):

            @self.training_logs.logs_button.click
            def show_logs():
                self.training_logs.toggle_logs()

        if self.hyperparameters_selector.run_model_benchmark_checkbox is not None:

            @self.hyperparameters_selector.run_model_benchmark_checkbox.value_changed
            def show_mb_speedtest(is_checked: bool):
                self.hyperparameters_selector.toggle_mb_speedtest(is_checked)

        # ------------------------------------------------- #

        self.layout: Widget = self.stepper

        # Run from experiment page

        train_task_id = getenv("modal.state.trainTaskId", None)
        if train_task_id is not None:
            train_task_id = int(train_task_id)
        train_mode = getenv("modal.state.trainMode", None)
        if train_task_id is not None and train_mode is not None:
            self._run_from_experiment(train_task_id, train_mode)
        # ----------------------------------------- #

    def set_next_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step + 1)

    def set_previous_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step - 1)

    def set_first_step(self):
        self.stepper.set_active_step(1)

    def set_last_step(self):
        self.stepper.set_active_step(len(self.steps))

    def enable_select_buttons(self):
        """
        Makes all select buttons in the GUI available for interaction.
        """
        if self.input_selector is not None:
            self.input_selector.button.enable()
        if self.model_selector is not None:
            self.model_selector.button.enable()
        if self.train_val_splits_selector is not None:
            self.train_val_splits_selector.button.enable()
        if self.classes_selector is not None:
            self.classes_selector.button.enable()
        if self.tags_selector is not None:
            self.tags_selector.button.enable()
        if self.model_selector is not None:
            self.model_selector.button.enable()
        if self.hyperparameters_selector is not None:
            self.hyperparameters_selector.button.enable()

    def disable_select_buttons(self):
        """
        Makes all select buttons in the GUI unavailable for interaction.
        """
        if self.input_selector is not None:
            self.input_selector.button.disable()
        if self.model_selector is not None:
            self.model_selector.button.disable()
        if self.train_val_splits_selector is not None:
            self.train_val_splits_selector.button.disable()
        if self.classes_selector is not None:
            self.classes_selector.button.disable()
        if self.tags_selector is not None:
            self.tags_selector.button.disable()
        if self.model_selector is not None:
            self.model_selector.button.disable()
        if self.hyperparameters_selector is not None:
            self.hyperparameters_selector.button.disable()

    # Set GUI from config
    def validate_app_state(self, app_state: dict) -> dict:
        """
        Validate the app state dictionary.

        :param app_state: The app state dictionary.
        :type app_state: dict
        """
        if not isinstance(app_state, dict):
            raise ValueError("app_state must be a dictionary")

        show_train_val = self.show_train_val_splits_selector
        show_classes = self.show_classes_selector
        show_tags = self.show_tags_selector

        # Basic required keys always needed
        base_required = {
            "model": ["source"],
        }
        if show_train_val:
            base_required["train_val_split"] = ["method"]
        if show_classes:
            base_required["classes"] = list
        if show_tags:
            base_required["tags"] = list

        for key, subkeys_or_type in base_required.items():
            if key not in app_state:
                raise KeyError(f"Missing required key in app_state: {key}")
            if isinstance(subkeys_or_type, list):
                for subkey in subkeys_or_type:
                    if subkey not in app_state[key]:
                        raise KeyError(f"Missing required key in app_state['{key}']: {subkey}")
            elif not isinstance(app_state[key], subkeys_or_type):
                valid_types = ""
                if isinstance(subkeys_or_type, tuple):
                    type_names = []
                    for t in subkeys_or_type:
                        if hasattr(t, "__name__"):
                            type_names.append(t.__name__)
                        else:
                            type_names.append(type(t).__name__)
                    valid_types = " or ".join(type_names)
                else:
                    if hasattr(subkeys_or_type, "__name__"):
                        valid_types = subkeys_or_type.__name__
                    else:
                        valid_types = type(subkeys_or_type).__name__
                raise ValueError(f"app_state['{key}'] must be of type {valid_types}")

        # Provide defaults for optional sections when selectors are disabled
        if not show_train_val:
            app_state.setdefault("train_val_split", {"method": "random"})
        if not show_classes:
            app_state.setdefault("classes", [])
        if not show_tags:
            app_state.setdefault("tags", [])

        model = app_state["model"]
        if model["source"] == "Pretrained models":
            if "model_name" not in model:
                raise KeyError("Missing required key in app_state['model']: model_name")
        elif model["source"] == "Custom models":
            custom_keys = ["task_id", "checkpoint"]
            for key in custom_keys:
                if key not in model:
                    raise KeyError(f"Missing required key in app_state['model']: {key}")

        options = app_state.setdefault(
            "options",
            {
                "model_benchmark": {
                    "enable": True,
                    "speed_test": True,
                },
                "cache_project": True,
            },
        )

        if not isinstance(options, dict):
            raise ValueError("app_state['options'] must be a dictionary")

        model_benchmark = options.setdefault(
            "model_benchmark", {"enable": True, "speed_test": True}
        )
        if not isinstance(model_benchmark, dict):
            raise ValueError("app_state['options']['model_benchmark'] must be a dictionary")
        model_benchmark.setdefault("enable", True)
        model_benchmark.setdefault("speed_test", True)

        if not isinstance(options.get("cache_project"), bool):
            raise ValueError("app_state['options']['cache_project'] must be a boolean")

        # Check train val splits
        train_val_splits_settings = app_state.get("train_val_split")
        if train_val_splits_settings.get("method") == "datasets":
            dataset_ids = []
            for parents, dataset in self._api.dataset.tree(self.project_id):
                dataset_ids.append(dataset.id)

            train_datasets = train_val_splits_settings.get("train_datasets", [])
            val_datasets = train_val_splits_settings.get("val_datasets", [])

            missing_datasets_ids = []
            for ds_id in train_datasets + val_datasets:
                if ds_id not in dataset_ids:
                    missing_datasets_ids.append(ds_id)

            if len(missing_datasets_ids) > 0:
                missing_datasets_text = ", ".join([str(ds_id) for ds_id in missing_datasets_ids])
                raise ValueError(
                    f"Datasets with ids: {missing_datasets_text} not found in the project"
                )
        elif train_val_splits_settings.get("method") == "tags":
            train_tag = train_val_splits_settings.get("train_tag")
            val_tag = train_val_splits_settings.get("val_tag")
            if not train_tag or not val_tag:
                raise ValueError("train_tag and val_tag must be specified in tags split method")
        elif train_val_splits_settings.get("method") == "random":
            split = train_val_splits_settings.get("split")
            percent = train_val_splits_settings.get("percent")
            if split not in ["train", "val"]:
                raise ValueError("split must be 'train' or 'val'")
            if not isinstance(percent, int) or not 0 < percent < 100:
                raise ValueError("percent must be an integer in range 1 to 99")
        elif train_val_splits_settings.get("method") == "collections":
            train_collections = train_val_splits_settings.get("train_collections", [])
            val_collections = train_val_splits_settings.get("val_collections", [])
            collection_ids = set()
            for collection in self._api.entities_collection.get_list(self.project_id):
                collection_ids.add(collection.id)
            missing_collections_ids = set(train_collections + val_collections) - collection_ids
            if missing_collections_ids:
                missing_collections_text = ", ".join(
                    [str(collection_id) for collection_id in missing_collections_ids]
                )
                raise ValueError(
                    f"Collections with ids: {missing_collections_text} not found in the project"
                )
        return app_state

    def load_from_app_state(
        self, app_state: Union[str, dict], click_cb: bool = True, validate_steps: bool = True
    ) -> None:
        """
        Load the GUI state from app state dictionary or path to the state file.

        :param app_state: The state dictionary or path to the state file.
        :type app_state: Union[str, dict]
        :param click_cb: Automatically click the callback functions to set the GUI state.
        :type click_cb: bool
        :param validate_steps: Validate the steps. If False, the steps will not be validated.
        :type validate_steps: bool

        app_state example:

            app_state = {
                "train_val_split": {
                    "method": "random",
                    "split": "train",
                    "percent": 90
                },
                "classes": ["apple"],
                # Pretrained model
                "model": {
                    "source": "Pretrained models",
                    "model_name": "rtdetr_r50vd_coco_objects365"
                },
                # Custom model
                # "model": {
                #     "source": "Custom models",
                #     "task_id": 555,
                #     "checkpoint": "checkpoint_10.pth"
                # },
                "hyperparameters": hyperparameters, # yaml string
                "options": {
                    "convert_class_shapes": True,
                    "model_benchmark": {
                        "enable": True,
                        "speed_test": True
                    },
                    "cache_project": True,
                    "export": {
                        "enable": True,
                        "ONNXRuntime": True,
                        "TensorRT": True
                    },
                },
                "experiment_name": "My Experiment",
                "start_training": False,
            }
        """
        if isinstance(app_state, str):
            if os.path.isfile(app_state):
                app_state = sly_json.load_json_file(app_state)
            else:
                app_state = json.loads(app_state)

        logger.debug("Loading from app state", extra={"app_state": app_state})
        app_state = self.validate_app_state(app_state)
        options = app_state.get("options", {})

        # Run init-steps and stop on validation failure
        def _run_step(init_fn, settings) -> bool:
            if not init_fn(settings, options, click_cb, validate_steps):
                return False
            return True

        # GUI init steps
        _steps = [
            (self._init_input, app_state.get("input"), "Input project"),
            (self._init_model, app_state["model"], "Select Model"),
            (self._init_classes, app_state.get("classes", []), "Classes Selector"),
            (self._init_tags, app_state.get("tags", []), "Tags Selector"),
            (self._init_train_val_splits, app_state.get("train_val_split", {}), "Train/Val Splits"),
            (self._init_hyperparameters, app_state.get("hyperparameters"), "Hyperparameters"),
        ]

        for idx, (init_fn, settings, step_name) in enumerate(_steps, start=1):
            if not _run_step(init_fn, settings):
                if validate_steps:
                    logger.warning(f"Step '{step_name}' {idx}/{len(_steps)} failed to validate")
                return
            if validate_steps:
                logger.info(
                    f"Step '{step_name}' {idx}/{len(_steps)} has been validated successfully"
                )

        # Set experiment name
        experiment_name = app_state.get("experiment_name")
        if experiment_name is not None and experiment_name != "":
            self.training_process.set_experiment_name(experiment_name)

        if validate_steps:
            logger.info(f"All steps have been validated successfully")

        self._start_training = app_state.get("start_training", False)
        # ------------------------------------------------------------------ #

    def _init_input(
        self,
        input_settings: Union[dict, None],
        options: dict,
        click_cb: bool = True,
        validate: bool = True,
    ) -> bool:
        """
        Initialize the input selector with the given settings.

        :param input_settings: The input settings.
        :type input_settings: dict
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """
        # Set Input
        self.input_selector.set_cache(options.get("cache_project", True))
        is_valid = True
        if validate:
            is_valid = self.input_selector.validate_step()
        if is_valid and click_cb:
            self.input_selector_cb()
            self.set_next_step()
        return is_valid
        # ----------------------------------------- #

    def _init_model(
        self,
        model_settings: dict,
        options: dict = None,
        click_cb: bool = True,
        validate: bool = True,
    ) -> bool:
        """
        Initialize the model selector with the given settings.

        :param model_settings: The model settings.
        :type model_settings: dict
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """

        # Pretrained
        if model_settings["source"] == ModelSource.PRETRAINED:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.PRETRAINED)
            self.model_selector.pretrained_models_table.set_by_model_name(
                model_settings["model_name"]
            )

        # Custom
        elif model_settings["source"] == ModelSource.CUSTOM:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.CUSTOM)
            self.model_selector.experiment_selector.set_selected_row_by_task_id(
                model_settings["task_id"]
            )
            experiment_info = self.model_selector.experiment_selector.get_selected_experiment_info()
            if model_settings["checkpoint"] not in experiment_info.checkpoints:
                if f"checkpoints/{model_settings['checkpoint']}" not in experiment_info.checkpoints:
                    raise ValueError(
                        f"Checkpoint '{model_settings['checkpoint']}' not found in selected task"
                    )
            self.model_selector.experiment_selector.set_selected_checkpoint_by_name(
                model_settings["checkpoint"]
            )

        is_valid = True
        if validate:
            is_valid = self.model_selector.validate_step()
        if is_valid and click_cb:
            self.model_selector_cb()
            self.set_next_step()
        return is_valid
        # ----------------------------------------- #

    def _init_classes(
        self, classes_settings: list, options: dict, click_cb: bool = True, validate: bool = True
    ) -> bool:
        """
        Initialize the classes selector with the given settings.

        :param classes_settings: The classes settings.
        :type classes_settings: list
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """
        if self.classes_selector is None:
            return True  # Selector disabled by app options

        convert_class_shapes = options.get("convert_class_shapes", True)
        if convert_class_shapes:
            self.classes_selector.convert_class_shapes_checkbox.check()

        # Set Classes
        if all(isinstance(c, int) for c in classes_settings):
            project_classes = []
            for obj_class in self.project_meta.obj_classes:
                if obj_class.sly_id in classes_settings:
                    project_classes.append(obj_class.name)
            classes_settings = project_classes

        self.classes_selector.set_classes(classes_settings)
        is_valid = True
        if validate:
            is_valid = self.classes_selector.validate_step()
        if is_valid and click_cb:
            self.classes_selector_cb()
            self.set_next_step()
        return is_valid
        # ----------------------------------------- #

    def _init_tags(
        self, tags_settings: list, options: dict, click_cb: bool = True, validate: bool = True
    ) -> bool:
        """
        Initialize the tags selector with the given settings.

        :param tags_settings: The tags settings.
        :type tags_settings: list
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """
        if self.tags_selector is None:
            return True  # Selector disabled by app options

        # Set Tags
        self.tags_selector.set_tags(tags_settings)
        is_valid = True
        if validate:
            is_valid = self.tags_selector.validate_step()
        if is_valid and click_cb:
            self.tags_selector_cb()
            self.set_next_step()
        return is_valid
        # ----------------------------------------- #

    def _init_train_val_splits(
        self,
        train_val_splits_settings: dict,
        options: dict,
        click_cb: bool = True,
        validate: bool = True,
    ) -> bool:
        """
        Initialize the train/val splits selector with the given settings.

        :param train_val_splits_settings: The train/val splits settings.
        :type train_val_splits_settings: dict
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """
        if self.train_val_splits_selector is None:
            return True  # Selector disabled by app options

        if train_val_splits_settings == {}:
            available_methods = self.app_options.get("train_val_splits_methods", [])
            if available_methods == []:
                method = "random"
                train_val_splits_settings = {"method": method, "split": "train", "percent": 80}
            else:
                method = available_methods[0]
                if method == "random":
                    train_val_splits_settings = {"method": method, "split": "train", "percent": 80}
                elif method == "tags":
                    train_val_splits_settings = {
                        "method": method,
                        "train_tag": "train",
                        "val_tag": "val",
                        "untagged_action": "ignore",
                    }
                elif method == "datasets":
                    train_val_splits_settings = {
                        "method": method,
                        "train_datasets": [],
                        "val_datasets": [],
                    }

        split_method = train_val_splits_settings["method"]
        if split_method == "random":
            split = train_val_splits_settings["split"]
            percent = train_val_splits_settings["percent"]
            self.train_val_splits_selector.train_val_splits.set_random_splits(split, percent)
        elif split_method == "tags":
            train_tag = train_val_splits_settings["train_tag"]
            val_tag = train_val_splits_settings["val_tag"]
            untagged_action = train_val_splits_settings["untagged_action"]
            self.train_val_splits_selector.train_val_splits.set_tags_splits(
                train_tag, val_tag, untagged_action
            )
        elif split_method == "datasets":
            train_datasets = train_val_splits_settings["train_datasets"]
            val_datasets = train_val_splits_settings["val_datasets"]
            self.train_val_splits_selector.train_val_splits.set_datasets_splits(
                train_datasets, val_datasets
            )
        elif split_method == "collections":
            train_collections = train_val_splits_settings["train_collections"]
            val_collections = train_val_splits_settings["val_collections"]
            self.train_val_splits_selector.train_val_splits.set_project_id_for_collections(
                self.project_id
            )
            self.train_val_splits_selector.train_val_splits.set_collections_splits(
                train_collections, val_collections
            )

        is_valid = True
        if validate:
            is_valid = self.train_val_splits_selector.validate_step()
        if is_valid and click_cb:
            self.train_val_splits_selector_cb()
            self.set_next_step()
        return is_valid

    def _init_hyperparameters(
        self,
        hyperparameters_settings: Union[dict, None],
        options: dict,
        click_cb: bool = True,
        validate: bool = True,
    ) -> bool:
        """
        Initialize the hyperparameters selector with the given settings.

        :param hyperparameters_settings: The hyperparameters settings.
        :type hyperparameters_settings: Union[dict, None]
        :param options: The application options.
        :type options: dict
        :param click_cb: Click the callback function.
        :type click_cb: bool
        :param validate: Validate the step.
        :type validate: bool
        """
        if hyperparameters_settings is not None:
            self.hyperparameters_selector.set_hyperparameters(hyperparameters_settings)

        model_benchmark_settings = options.get("model_benchmark", None)
        if model_benchmark_settings is not None:
            self.hyperparameters_selector.set_model_benchmark_checkbox_value(
                model_benchmark_settings["enable"]
            )
            self.hyperparameters_selector.set_speedtest_checkbox_value(
                model_benchmark_settings["speed_test"]
                if model_benchmark_settings["enable"]
                else False
            )
        export_weights_settings = options.get("export", None)
        if export_weights_settings is not None:
            self.hyperparameters_selector.set_export_onnx_checkbox_value(
                export_weights_settings.get(RuntimeType.ONNXRUNTIME, False)
            )
            self.hyperparameters_selector.set_export_tensorrt_checkbox_value(
                export_weights_settings.get(RuntimeType.TENSORRT, False)
            )

        is_valid = True
        if validate:
            is_valid = self.hyperparameters_selector.validate_step()
        if is_valid and click_cb:
            self.hyperparameters_selector_cb()
            self.set_next_step()
        return is_valid

    # ----------------------------------------- #

    # Run from experiment page
    def _download_experiment_state(self, experiment_info: ExperimentInfo) -> dict:
        local_app_state_path = f"./app_state.json"
        remote_app_state_path = os.path.join(experiment_info.artifacts_dir, "app_state.json")
        self._api.file.download(self.team_id, remote_app_state_path, local_app_state_path)
        app_state = sly_json.load_json_file(local_app_state_path)
        sly_fs.silent_remove(local_app_state_path)
        return app_state

    def _download_experiment_hparams(self, experiment_info: ExperimentInfo) -> dict:
        local_hparams_path = f"./{experiment_info.hyperparameters}"
        remote_hparams_path = os.path.join(
            experiment_info.artifacts_dir, experiment_info.hyperparameters
        )
        self._api.file.download(self.team_id, remote_hparams_path, local_hparams_path)
        with open(local_hparams_path, "r") as f:
            hparams = f.read()
        sly_fs.silent_remove(local_hparams_path)
        return hparams

    def _run_from_experiment(self, train_task_id: int, train_mode: str):
        experiment_info = self._api.nn.get_experiment_info(train_task_id)
        experiment_state = experiment_info.app_state

        if train_mode == "continue":
            model_settings = {
                "source": ModelSource.CUSTOM,
                "task_id": train_task_id,
                "checkpoint": experiment_info.best_checkpoint,
            }

        if experiment_state is not None:
            self.input_selector.validator_text.set(
                f"Training configuration is loaded from the experiment: {experiment_info.experiment_name}.",
                "success",
            )
            self.input_selector.validator_text.show()
            experiment_state = self._download_experiment_state(experiment_info)
            if train_mode == "continue":
                experiment_state["model"] = model_settings
            self.load_from_app_state(experiment_state, click_cb=False, validate_steps=False)
        else:
            self.input_selector.validator_text.set(
                f"Couldn't load full training configuration from the experiment: {experiment_info.experiment_name}. Only model and hyperparameters are loaded.",
                "warning",
            )
            self.input_selector.validator_text.show()
            hparams = self._download_experiment_hparams(experiment_info)
            self.hyperparameters_selector.set_hyperparameters(hparams)
            if train_mode == "continue":
                self._init_model(model_settings, {}, click_cb=False, validate=False)
    # ----------------------------------------- #

    def _extract_state_from_env(self):
        import ast
        import os

        base = "modal.state"
        state = {}
        for key, value in os.environ.items():
            state_part = state
            if key.startswith(base):
                key = key.replace(base + ".", "")
                parts = key.split(".")
                while len(parts) > 1:
                    part = parts.pop(0)
                    state_part.setdefault(part, {})
                    state_part = state_part[part]
                part = parts.pop(0)
                if value and (value[0] == "[" or value.isdigit()):
                    state_part[part] = ast.literal_eval(value)
                elif value in ["True", "true", "False", "false"]:
                    state_part[part] = value in ["True", "true"]
                else:
                    state_part[part] = value
        return state
    # ----------------------------------------- #



import os
import shutil
import time

import cv2
import ffmpeg

import supervisely as sly
from supervisely.io.fs import ensure_base_path
from tqdm import tqdm


class InferenceVideoInterface:
    """Interface for video inference: downloads frames and manages frame indexes for batch processing."""

    def __init__(
        self,
        api,
        start_frame_index,
        frames_count,
        frames_direction,
        video_info,
        imgs_dir,
        preparing_progress,
    ):
        """
        :param api: Supervisely API.
        :type api: Api
        :param start_frame_index: Start frame.
        :type start_frame_index: int
        :param frames_count: Number of frames.
        :type frames_count: int
        :param frames_direction: 'forward' or 'backward'.
        :type frames_direction: str
        :param video_info: Video metadata.
        :type video_info: VideoInfo
        :param imgs_dir: Temp directory for frames.
        :type imgs_dir: str
        :param preparing_progress: Progress callback.
        :type preparing_progress: Callable
        """
        self.api = api

        self.video_info = video_info
        self.images_paths = []

        self.start_frame_index = start_frame_index
        self.frames_count = frames_count
        self.frames_direction = frames_direction

        self._video_fps = round(1 / self.video_info.frames_to_timecodes[1])

        self._geometries = []
        self._frames_indexes = []

        self._add_frames_indexes()

        self._frames_path = os.path.join(
            imgs_dir, f"video_inference_{video_info.id}_{time.time_ns()}", "frames"
        )
        self._imgs_dir = imgs_dir

        self._preparing_progress = preparing_progress

        self._local_video_path = None

        os.makedirs(self._frames_path, exist_ok=True)

    def _add_frames_indexes(self):
        total_frames = self.video_info.frames_count
        cur_index = self.start_frame_index

        if self.frames_direction == "forward":
            end_point = (
                cur_index + self.frames_count
                if cur_index + self.frames_count < total_frames
                else total_frames
            )
            self._frames_indexes = [
                curr_frame_index for curr_frame_index in range(cur_index, end_point, 1)
            ]
        else:
            end_point = cur_index - self.frames_count if cur_index - self.frames_count > -1 else -1
            self._frames_indexes = [
                curr_frame_index for curr_frame_index in range(cur_index, end_point, -1)
            ]
            self._frames_indexes = []

    def _download_video_by_frames(self):
        self._preparing_progress["status"] = "download_frames"
        self._preparing_progress["current"] = 0

        def download_frame(idx, frame_index):
            frame_path = os.path.join(f"{self._frames_path}", f"frame{idx:06d}.png")
            if os.path.isfile(frame_path):
                return
            img_rgb = self.api.video.frame.download_np(self.video_info.id, frame_index)
            sly.image.write(frame_path, img_rgb)
            self.images_paths.append(frame_path)
            self._preparing_progress["current"] += 1

        if len(self._frames_indexes) == 1:
            download_frame(0, self._frames_indexes[0])
            return

        for index, frame_index in tqdm(
            enumerate(self._frames_indexes),
            desc="Downloading frames",
            total=len(self._frames_indexes),
        ):
            download_frame(index, frame_index)

    def _download_entire_video(self):
        def videos_to_frames(video_path, progress, frames_range=None):
            progress["status"] = "cut"
            progress["current"] = 0
            progress["total"] = self.frames_count
            vidcap = cv2.VideoCapture(video_path)
            vidcap.set(cv2.CAP_PROP_ORIENTATION_AUTO, 1)
            success, image = vidcap.read()
            count = 0
            while success:
                output_image_path = os.path.join(f"{self._frames_path}", f"frame{count:06d}.png")
                if frames_range:
                    if frames_range[0] <= count <= frames_range[1]:
                        cv2.imwrite(output_image_path, image)  # save frame as PNG file
                        self.images_paths.append(output_image_path)
                else:
                    cv2.imwrite(output_image_path, image)  # save frame as PNG file
                    self.images_paths.append(output_image_path)
                success, image = vidcap.read()

                count += 1
                progress["current"] = count

            fps = vidcap.get(cv2.CAP_PROP_FPS)

            return {"frames_path": self._frames_path, "fps": fps, "video_path": video_path}

        def download_video(video_path, video_info, progress):
            progress["status"] = "download_video"
            progress["current"] = 0
            progress["total"] = int(video_info.file_meta["size"])
            response = self.api.video._download(video_info.id, is_stream=True)
            ensure_base_path(video_path)
            with open(video_path, "wb") as fd:
                mb1 = 1024 * 1024
                for chunk in response.iter_content(chunk_size=mb1):
                    fd.write(chunk)
                    progress["current"] += len(chunk)

        self._local_video_path = os.path.join(
            self._imgs_dir, f"{time.time_ns()}_{self.video_info.name}"
        )
        download_video(self._local_video_path, self.video_info, self._preparing_progress)
        return videos_to_frames(
            self._local_video_path,
            self._preparing_progress,
            [self.start_frame_index, self.start_frame_index + self.frames_count - 1],
        )

    def download_frames(self):
        if self.frames_count > (self.video_info.frames_count * 0.3):
            sly.logger.debug("Downloading entire video")
            self._download_entire_video()
        else:
            sly.logger.debug("Downloading video frame by frame")
            self._download_video_by_frames()
        self._preparing_progress["status"] = "inference"

    def __del__(self):
        if os.path.isdir(self._frames_path):
            shutil.rmtree(os.path.dirname(self._frames_path), ignore_errors=True)

            if self._local_video_path is not None and os.path.isfile(self._local_video_path):
                os.remove(self._local_video_path)



import shutil
import tempfile
from collections import defaultdict
from pathlib import Path
from typing import Iterator, List, Optional, Tuple, Union

import cv2
import ffmpeg
import numpy as np

from supervisely import VideoAnnotation, logger
from supervisely.geometry.geometry import Geometry
from supervisely.nn.model.prediction import Prediction
from supervisely.nn.tracker.utils import predictions_to_video_annotation


class TrackingVisualizer:
    """Renders tracking results onto video: boxes, trajectories, labels, frame numbers."""

    def __init__(
        self,
        show_labels: bool = True,
        show_classes: bool = True,
        show_trajectories: bool = True,
        show_frame_number: bool = False,
        box_thickness: int = 2,
        text_scale: float = 0.6,
        text_thickness: int = 2,
        trajectory_length: int = 30,
        codec: str = "mp4",
        output_fps: float = 30.0,
        colorize_tracks: bool = True,
        trajectory_thickness: int = 2,
    ):
        """
        Initialize the visualizer with configuration.

        :param show_labels: Whether to show track IDs.
        :type show_labels: bool
        :param show_classes: Whether to show class names.
        :type show_classes: bool
        :param show_trajectories: Whether to draw trajectories.
        :type show_trajectories: bool
        :param show_frame_number: Whether to overlay frame number.
        :type show_frame_number: bool
        :param box_thickness: Thickness of bounding boxes.
        :type box_thickness: int
        :param text_scale: Scale of label text.
        :type text_scale: float
        :param text_thickness: Thickness of label text.
        :type text_thickness: int
        :param trajectory_length: How many points to keep in trajectory.
        :type trajectory_length: int
        :param codec: Output video codec.
        :type codec: str
        :param output_fps: Output video framerate.
        :type output_fps: float
        :param colorize_tracks: If True, ignore colors from project meta and generate new colors for each tracked object; if False, try to use colors from project meta when possible.
        :type colorize_tracks: bool
        :param trajectory_thickness: Thickness of trajectory lines.
        :type trajectory_thickness: int
        :returns: None
        :rtype: None
        """
        # Visualization settings
        self.show_labels = show_labels
        self.show_classes = show_classes
        self.show_trajectories = show_trajectories
        self.show_frame_number = show_frame_number

        # Style settings
        self.box_thickness = box_thickness
        self.text_scale = text_scale
        self.text_thickness = text_thickness
        self.trajectory_length = trajectory_length
        self.trajectory_thickness = trajectory_thickness
        self.colorize_tracks = colorize_tracks

        # Output settings
        self.codec = codec
        self.output_fps = output_fps

        # Internal state
        self.annotation = None
        self.tracks_by_frame = {}
        self.track_centers = defaultdict(list)
        self.track_colors = {}
        self.color_palette = self._generate_color_palette()
        self._temp_dir = None

    def _generate_color_palette(self, num_colors: int = 100) -> List[Tuple[int, int, int]]:
        """
        Generate bright, distinct color palette for track visualization.
        Uses HSV space with random hue and fixed high saturation/value.
        """
        np.random.seed(42)
        colors = []
        for i in range(num_colors):
            hue = np.random.randint(0, 180)
            saturation = 200 + np.random.randint(55)
            value = 200 + np.random.randint(55)

            hsv_color = np.uint8([[[hue, saturation, value]]])
            bgr_color = cv2.cvtColor(hsv_color, cv2.COLOR_HSV2BGR)[0][0]
            colors.append(tuple(map(int, bgr_color)))
        return colors

    def _get_track_color(self, track_id: int) -> Tuple[int, int, int]:
        """Get consistent color for track ID from palette."""
        return self.color_palette[track_id % len(self.color_palette)]

    def _get_video_info(self, video_path: Path) -> Tuple[int, int, float, int]:
        """
        Get video metadata using ffmpeg.
        
        Returns:
            Tuple of (width, height, fps, total_frames)
        """
        try:
            probe = ffmpeg.probe(str(video_path))
            video_stream = next((stream for stream in probe['streams'] 
                               if stream['codec_type'] == 'video'), None)

            if video_stream is None:
                raise ValueError(f"No video stream found in: {video_path}")

            width = int(video_stream['width'])
            height = int(video_stream['height'])

            # Extract FPS
            fps_str = video_stream.get('r_frame_rate', '30/1')
            if '/' in fps_str:
                num, den = map(int, fps_str.split('/'))
                fps = num / den if den != 0 else 30.0
            else:
                fps = float(fps_str)

            # Get total frames
            total_frames = int(video_stream.get('nb_frames', 0))
            if total_frames == 0:
                # Fallback: estimate from duration and fps
                duration = float(video_stream.get('duration', 0))
                total_frames = int(duration * fps) if duration > 0 else 0

            return width, height, fps, total_frames

        except Exception as e:
            raise ValueError(f"Could not read video metadata {video_path}: {str(e)}")

    def _create_frame_iterator(self, source: Union[str, Path]) -> Iterator[Tuple[int, np.ndarray]]:
        """
        Create iterator that yields (frame_index, frame) tuples.

        :param source: Path to video file or directory with frame images
        :type source: Union[str, Path]
        :returns: Iterator that yields (frame_index, frame) tuples
        :rtype: Iterator[Tuple[int, np.ndarray]]
        """
        source = Path(source)

        if source.is_file():
            yield from self._iterate_video_frames(source)
        elif source.is_dir():
            yield from self._iterate_directory_frames(source)
        else:
            raise ValueError(f"Source must be a video file or directory, got: {source}")

    def _iterate_video_frames(self, video_path: Path) -> Iterator[Tuple[int, np.ndarray]]:
        """
        Iterate through video frames using ffmpeg.

        :param video_path: Path to video file
        :type video_path: Path
        :returns: Iterator that yields (frame_index, frame) tuples
        :rtype: Iterator[Tuple[int, np.ndarray]]
        """
        width, height, fps, total_frames = self._get_video_info(video_path)

        # Store video info for later use
        self.source_fps = fps
        self.frame_size = (width, height)

        process = (
            ffmpeg
            .input(str(video_path))
            .output('pipe:', format='rawvideo', pix_fmt='bgr24', loglevel='quiet')
            .run_async(pipe_stdout=True, pipe_stderr=False)
        )

        try:
            frame_size_bytes = width * height * 3
            frame_idx = 0

            while True:
                frame_data = process.stdout.read(frame_size_bytes)
                if len(frame_data) != frame_size_bytes:
                    break

                frame = np.frombuffer(frame_data, np.uint8).reshape([height, width, 3])
                yield frame_idx, frame
                frame_idx += 1

        except ffmpeg.Error as e:
            logger.error(f"ffmpeg error: {e.stderr.decode() if e.stderr else str(e)}", exc_info=True)

        finally:
            process.stdout.close()
            if process.stderr:
                process.stderr.close()
            process.wait()

    def _iterate_directory_frames(self, frames_dir: Path) -> Iterator[Tuple[int, np.ndarray]]:
        """Iterate through image frames in directory."""
        if not frames_dir.is_dir():
            raise ValueError(f"Directory does not exist: {frames_dir}")

        # Support common image extensions
        extensions = ['.jpg', '.jpeg', '.png', '.bmp', '.tiff']
        image_files = []
        for ext in extensions:
            image_files.extend(frames_dir.glob(f'*{ext}'))
            image_files.extend(frames_dir.glob(f'*{ext.upper()}'))

        image_files = sorted(image_files)
        if not image_files:
            raise ValueError(f"No image files found in directory: {frames_dir}")

        # Set fps from config for image sequences
        self.source_fps = self.output_fps

        for frame_idx, img_path in enumerate(image_files):
            frame = cv2.imread(str(img_path))
            if frame is not None:
                if frame_idx == 0:
                    h, w = frame.shape[:2]
                    self.frame_size = (w, h)
                yield frame_idx, frame
            else:
                logger.warning(f"Could not read image: {img_path}")

    def _extract_tracks_from_annotation(self) -> None:
        """
        Extract tracking data from Supervisely VideoAnnotation.
        
        Populates self.tracks_by_frame with frame-indexed tracking data.
        """
        self.tracks_by_frame = defaultdict(list)
        self.track_colors = {}

        # Map object keys to track info
        objects = {}
        for i, obj in enumerate(self.annotation.objects):
            objects[obj.key] = (i, obj.obj_class.name)

        # Extract tracks from frames
        for frame in self.annotation.frames:
            frame_idx = frame.index
            for figure in frame.figures:
                object_key = figure.parent_object.key
                if object_key not in objects:
                    continue

                track_id, class_name = objects[object_key]

                if track_id not in self.track_colors:
                    if self.colorize_tracks:
                        # auto-color override everything
                        color = self._get_track_color(track_id)
                    else:
                        # try to use annotation color
                        color = figure.video_object.obj_class.color
                        if color:
                            # convert rgb → bgr
                            color = color[::-1]
                        else:
                            # fallback to auto-color if annotation missing
                            color = self._get_track_color(track_id)

                    self.track_colors[track_id] = color

                self.tracks_by_frame[frame_idx].append((track_id, figure.geometry, class_name))

        logger.info(f"Extracted tracks from {len(self.tracks_by_frame)} frames")

    def _draw_detection(
        self, img: np.ndarray, track_id: int, geometry: Geometry, class_name: str
    ) -> Optional[Tuple[int, int]]:
        """
        Draw single detection with track ID and class label.
        Returns the center point of the bbox for trajectory drawing.
        """
        rect = geometry.to_bbox()
        bbox = (rect.left, rect.top, rect.right, rect.bottom)

        x1, y1, x2, y2 = map(int, bbox)

        if x2 <= x1 or y2 <= y1:
            return None

        color = self.track_colors[track_id]

        # Draw bounding box
        geometry.draw_contour(img, color=color, thickness=self.box_thickness)
        # cv2.rectangle(img, (x1, y1), (x2, y2), color, self.box_thickness)

        # Draw label if enabled
        if self.show_labels:
            label = f"ID:{track_id}"
            if self.show_classes:
                label += f" ({class_name})"

            label_y = y1 - 10 if y1 > 30 else y2 + 25
            (text_w, text_h), _ = cv2.getTextSize(
                label, cv2.FONT_HERSHEY_SIMPLEX, self.text_scale, self.text_thickness
            )

            cv2.rectangle(img, (x1, label_y - text_h - 5), 
                        (x1 + text_w, label_y + 5), color, -1)
            cv2.putText(img, label, (x1, label_y), 
                    cv2.FONT_HERSHEY_SIMPLEX, self.text_scale, 
                    (255, 255, 255), self.text_thickness, cv2.LINE_AA)

        # Return center point for trajectory
        return (x1 + x2) // 2, (y1 + y2) // 2

    def _draw_trajectories(self, img: np.ndarray) -> None:
        """Draw trajectory lines for all tracks, filtering out big jumps."""
        if not self.show_trajectories:
            return

        max_jump = 200  

        for centers_with_colors in self.track_centers.values():

            if len(centers_with_colors) < 2:
                continue

            points, colors = zip(*centers_with_colors[-self.trajectory_length :])

            for i in range(1, len(points)):
                color = colors[i]
                p1, p2 = points[i - 1], points[i]
                if p1 is None or p2 is None:
                    continue

                if np.hypot(p2[0] - p1[0], p2[1] - p1[1]) > max_jump:
                    continue
                cv2.line(img, p1, p2, color, self.trajectory_thickness)
                cv2.circle(img, p1, 3, color, -1)

    def _process_single_frame(self, frame: np.ndarray, frame_idx: int) -> np.ndarray:
        """
        Process single frame: add annotations and return processed frame.

        :param frame: Input frame
        :type frame: np.ndarray
        :param frame_idx: Frame index
        :type frame_idx: int
        :returns: Annotated frame
        :rtype: np.ndarray
        """
        img = frame.copy()
        active_ids = set()
        # Draw detections for current frame
        if frame_idx in self.tracks_by_frame:
            for track_id, geometry, class_name in self.tracks_by_frame[frame_idx]:
                center = self._draw_detection(img, track_id, geometry, class_name)
                color = self.track_colors[track_id]
                self.track_centers[track_id].append((center, color))
                if len(self.track_centers[track_id]) > self.trajectory_length:
                    self.track_centers[track_id].pop(0)
                active_ids.add(track_id)

        for tid in self.track_centers.keys():
            if tid not in active_ids:
                self.track_centers[tid].append((None, None))
                if len(self.track_centers[tid]) > self.trajectory_length:
                    self.track_centers[tid].pop(0)

        # Draw trajectories
        self._draw_trajectories(img)

        # Add frame number if requested
        if self.show_frame_number:
            cv2.putText(img, f"Frame: {frame_idx + 1}", (10, 30),
                       cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 255, 255), 2, cv2.LINE_AA)

        return img

    def _save_processed_frame(self, frame: np.ndarray, frame_idx: int) -> str:
        """
        Save processed frame to temporary directory.

        :param frame: Processed frame
        :type frame: np.ndarray
        :param frame_idx: Frame index
        :type frame_idx: int
        :returns: Path to saved frame
        :rtype: str
        """
        frame_path = self._temp_dir / f"frame_{frame_idx:08d}.jpg"
        cv2.imwrite(str(frame_path), frame, [cv2.IMWRITE_JPEG_QUALITY, 95])
        return str(frame_path)

    def _create_video_from_frames(self, output_path: Union[str, Path]) -> None:
        """
        Create final video from processed frames using ffmpeg.

        :param output_path: Path for output video
        :type output_path: Union[str, Path]
        """
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        # Create video from frame sequence
        input_pattern = str(self._temp_dir / "frame_%08d.jpg")

        try:
            (
                ffmpeg
                .input(input_pattern, pattern_type='sequence', framerate=self.source_fps)
                .output(str(output_path), vcodec='libx264', pix_fmt='yuv420p', crf=18)
                .overwrite_output()
                .run(capture_stdout=True, capture_stderr=True)
            )
            logger.info(f"Video saved to {output_path}")

        except ffmpeg.Error as e:
            error_msg = e.stderr.decode() if e.stderr else "Unknown ffmpeg error"
            raise ValueError(f"Failed to create video: {error_msg}")

    def _cleanup_temp_directory(self) -> None:
        """Clean up temporary directory and all its contents."""
        if self._temp_dir and self._temp_dir.exists():
            shutil.rmtree(self._temp_dir)
            self._temp_dir = None

    def visualize_video_annotation(self, annotation: VideoAnnotation, 
                                  source: Union[str, Path], 
                                  output_path: Union[str, Path]) -> None:
        """
        Visualize tracking annotations on video using streaming approach.

        :param annotation: Supervisely VideoAnnotation object with tracking data
        :type annotation: VideoAnnotation
        :param source: Path to video file or directory containing frame images
        :type source: Union[str, Path]
        :param output_path: Path for output video file
        :type output_path: Union[str, Path]
        :returns: None
        :rtype: None
        """
        if not isinstance(annotation, VideoAnnotation):
            raise TypeError(f"Annotation must be VideoAnnotation, got {type(annotation)}")

        # Store annotation
        self.annotation = annotation

        # Create temporary directory for processed frames
        self._temp_dir = Path(tempfile.mkdtemp(prefix="video_viz_"))

        try:
            # Extract tracking data
            self._extract_tracks_from_annotation()

            if not self.tracks_by_frame:
                logger.warning("No tracking data found in annotation")

            # Reset trajectory tracking
            self.track_centers = defaultdict(list)

            # Process frames one by one
            frame_count = 0
            for frame_idx, frame in self._create_frame_iterator(source):
                # Process frame
                processed_frame = self._process_single_frame(frame, frame_idx)

                # Save processed frame
                self._save_processed_frame(processed_frame, frame_idx)

                frame_count += 1

                # Progress logging
                if frame_count % 100 == 0:
                    logger.info(f"Processed {frame_count} frames")

            logger.info(f"Finished processing {frame_count} frames")

            # Create final video from saved frames
            self._create_video_from_frames(output_path)

        finally:
            # Always cleanup temporary files
            self._cleanup_temp_directory()

    def __del__(self):
        """Cleanup temporary directory on object destruction."""
        self._cleanup_temp_directory()


def visualize(
    predictions: Union[VideoAnnotation, List[Prediction]], 
    source: Union[str, Path], 
    output_path: Union[str, Path],
    show_labels: bool = True,
    show_classes: bool = True,
    show_trajectories: bool = True,
    box_thickness: int = 2,
    colorize_tracks: bool = True,
    **kwargs
) -> None:
    """
    Visualize tracking results from either VideoAnnotation or list of Prediction.

    :param predictions: Tracking data to render; either a Supervisely VideoAnnotation or a list of Prediction objects.
    :type predictions: Union[VideoAnnotation, List[Prediction]]
    :param source: Path to an input video file or a directory of sequential frames (e.g., frame_000001.jpg).
    :type source: Union[str, Path]
    :param output_path: Path to the output video file to be created.
    :type output_path: Union[str, Path]
    :param show_labels: Draw per-object labels (track IDs).
    :type show_labels: bool
    :param show_classes: Draw class names for each object.
    :type show_classes: bool
    :param show_trajectories: Render object trajectories across frames.
    :type show_trajectories: bool
    :param box_thickness: Bounding-box line thickness in pixels.
    :type box_thickness: int
    :param colorize_tracks: if True, ignore colors from project meta and generate new colors for each tracked object; if False, try to use colors from project meta when possible.
    :type colorize_tracks: bool
    :param kwargs: Additional keyword arguments for TrackingVisualizer.
    :type kwargs: dict
    :returns: None
    :rtype: None
    """
    visualizer = TrackingVisualizer(
        show_labels=show_labels, 
        show_classes=show_classes, 
        show_trajectories=show_trajectories,
        box_thickness=box_thickness,
        colorize_tracks=colorize_tracks,
        **kwargs
    )

    if isinstance(predictions, VideoAnnotation):
        visualizer.visualize_video_annotation(predictions, source, output_path)
    elif isinstance(predictions, list):
        predictions = predictions_to_video_annotation(predictions)
        visualizer.visualize_video_annotation(predictions, source, output_path)
    else:
        raise TypeError(f"Predictions must be VideoAnnotation or list of Prediction, got {type(predictions)}")



from typing import Dict, List, Literal, Union

from supervisely.api.api import Api
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.io.fs import get_file_ext
from supervisely.nn.utils import ModelSource, _get_model_name


class PretrainedModelsSelector(Widget):
    """Widget for selecting pretrained models from a list. Groups models by architecture and task type."""

    class Routes:
        """Route name constants for this widget."""
        ARCH_TYPE_CHANGED = "arch_type_changed"
        TASK_TYPE_CHANGED = "task_type_changed"
        MODEL_CHANGED = "model_changed"

    def __init__(
        self,
        models_list: List[Dict],
        widget_id: str = None,
        sort_models: bool = False,
    ):
        """Initialize the PretrainedModelsSelector widget.

        :param models_list: List of models to display. See usage example for expected format.
        :type models_list: List[Dict]
        :param widget_id: Unique identifier for the widget.
        :type widget_id: str, optional
        :param sort_models: Whether to sort task types within each architecture by name.
        :type sort_models: bool, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import PretrainedModelsSelector

                models_list = [
                    {"Model": "YOLOv8n-det", "meta": {"taskType": "object detection", "weightsURL": "https://..."}},
                    {"Model": "YOLOv8s-det", "meta": {"taskType": "object detection", "weightsURL": "https://..."}},
                ]
                pretrained_models_selector = PretrainedModelsSelector(models_list=models_list, sort_models=False)
        """
        self._api = Api.from_env()

        self._models = models_list
        filtered_models = self._filter_and_sort_models(self._models, sort_models)

        self._table_data = filtered_models
        self._task_types = self._filter_task_types(list(filtered_models.keys()))
        self._arch_types = []
        # maintain correct order of arch types
        for task_type in self._task_types:
            for arch_type in filtered_models[task_type].keys():
                if arch_type not in self._arch_types:
                    self._arch_types.append(arch_type)

        self._arch_changes_handled = False
        self._task_changes_handled = False
        self._model_changes_handled = False

        self.__default_selected_arch_type = (
            self._arch_types[0] if len(self._arch_types) > 0 else None
        )
        self.__default_selected_task_type = (
            self._task_types[0] if len(self._task_types) > 0 else None
        )

        super().__init__(widget_id=widget_id, file_path=__file__)

    @property
    def models(self) -> List[str]:
        return self._models

    def get_json_data(self) -> Dict:
        return {
            "tableData": self._table_data,
        }

    def get_json_state(self) -> Dict:
        return {
            "selectedRow": 0,
            "taskTypes": self._task_types,
            "archTypes": self._arch_types,
            "selectedTaskType": self.__default_selected_task_type,
            "selectedArchType": self.__default_selected_arch_type,
        }

    def get_available_task_types(self) -> List[str]:
        return StateJson()[self.widget_id]["taskTypes"]

    def get_available_arch_types(self) -> List[str]:
        return StateJson()[self.widget_id]["archTypes"]

    def get_selected_task_type(self) -> str:
        return StateJson()[self.widget_id]["selectedTaskType"]

    def get_selected_arch_type(self) -> str:
        return StateJson()[self.widget_id]["selectedArchType"]

    def get_selected_row(self, state=StateJson()) -> Union[Dict, None]:
        task_type = self.get_selected_task_type()
        arch_type = self.get_selected_arch_type()
        if task_type is None or arch_type is None:
            return

        models = self._table_data[task_type][arch_type]
        if len(models) == 0:
            return
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        if widget_actual_state is not None and widget_actual_data is not None:
            selected_row_index = int(widget_actual_state["selectedRow"])
            return models[selected_row_index]

    def get_selected_model_params(
        self,
        model_name_column: str = "Model",
        train_version: Literal["v1", "v2"] = "v1",
    ) -> Union[Dict, None]:
        selected_model = self.get_selected_row()
        if selected_model is None:
            return {}

        if train_version == "v1":
            model_name = selected_model.get(model_name_column)
            if model_name is None:
                model_name = _get_model_name(selected_model)
                if model_name is None:
                    raise ValueError(
                        "Could not find model name. Make sure you have column 'Model' in your models list."
                    )

            model_meta = selected_model.get("meta")
            if model_meta is None:
                raise ValueError(
                    "Could not find model meta. Make sure you have key 'meta' in your models configuration list."
                )
            checkpoint_url = model_meta.get("weights_url")
            if checkpoint_url is None:
                model_files = model_meta.get("model_files")
                if model_files is None:
                    raise ValueError(
                        "Could not find model files. Make sure you have key 'model_files' or 'weights_url' in 'meta' in your models configuration list."
                    )
                checkpoint_url = model_files.get("checkpoint")
                if checkpoint_url is None:
                    raise ValueError(
                        "Could not find checkpoint url. Make sure you have key 'checkpoint' in 'model_files' in 'meta' in your models configuration list."
                    )

            checkpoint_ext = get_file_ext(checkpoint_url)
            checkpoint_name = f"{model_name.lower()}{checkpoint_ext}"

            task_type = self.get_selected_task_type()
            model_params = {
                "model_source": "Pretrained models",
                "task_type": task_type,
                "checkpoint_name": checkpoint_name,
                "checkpoint_url": checkpoint_url,
            }

            if len(self._arch_types) > 1:
                arch_type = self.get_selected_arch_type()
                model_params["arch_type"] = arch_type

            config_url = selected_model.get("meta", {}).get("config_url")
            if config_url is not None:
                model_params["config_url"] = config_url
        elif train_version == "v2":
            model_info = self.get_selected_row()
            meta = model_info.get("meta")
            if meta is None:
                raise ValueError("key 'meta' not found in model configuration")
            model_files = meta.get("model_files")
            if model_files is None:
                raise ValueError("key 'model_files' not found in key 'meta' in model configuration")
            model_params = {
                "model_source": ModelSource.PRETRAINED,
                "model_info": model_info,
                "model_files": model_files,
            }
        return model_params

    def get_selected_row_index(self, state=StateJson()) -> Union[int, None]:
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        if widget_actual_state is not None and widget_actual_data is not None:
            return widget_actual_state["selectedRow"]

    def set_active_arch_type(self, arch_type: str) -> None:
        if arch_type not in self._arch_types:
            raise ValueError(f'Architecture type "{arch_type}" does not exist')
        StateJson()[self.widget_id]["selectedArchType"] = arch_type
        StateJson().send_changes()

    def set_active_task_type(self, task_type: str) -> None:
        if task_type not in self._task_types:
            raise ValueError(f'Task type "{task_type}" does not exist')
        StateJson()[self.widget_id]["selectedTaskType"] = task_type
        StateJson().send_changes()

    def set_active_row(self, row_index: int) -> None:
        if row_index < 0:
            raise ValueError(f'Row with index "{row_index}" does not exist')
        StateJson()[self.widget_id]["selectedRow"] = row_index
        StateJson().send_changes()

    def set_by_model_name(self, model_name: str) -> None:
        for task_type in self._table_data:
            for arch_type in self._table_data[task_type]:
                for idx, model in enumerate(self._table_data[task_type][arch_type]):
                    name_from_info = _get_model_name(model)
                    if name_from_info is not None:
                        if name_from_info.lower() == model_name.lower():
                            self.set_active_task_type(task_type)
                            self.set_active_arch_type(arch_type)
                            self.set_active_row(idx)
                            return

    def get_by_model_name(self, model_name: str) -> Union[Dict, None]:
        for task_type in self._table_data:
            for arch_type in self._table_data[task_type]:
                for idx, model in enumerate(self._table_data[task_type][arch_type]):
                    name_from_info = _get_model_name(model)
                    if name_from_info is not None:
                        if name_from_info.lower() == model_name.lower():
                            return model

    def _filter_and_sort_models(self, models: List[Dict], sort_models: bool = True) -> Dict:
        filtered_models = {}
        for model in models:
            for key in model:
                if isinstance(model[key], (int, float)):
                    model[key] = str(model[key])

            arch_type = model.get("meta", {}).get("arch_type", "other")
            task_type = model.get("meta", {}).get("task_type", model.get("task_type", "other"))

            if task_type not in filtered_models:
                filtered_models[task_type] = {}
            if arch_type not in filtered_models[task_type]:
                filtered_models[task_type][arch_type] = []
            filtered_models[task_type][arch_type].append(model)

        if sort_models:
            sorted_filtered_models = {
                task: {arch: models for arch, models in sorted(archs.items())}
                for task, archs in sorted(filtered_models.items())
            }
        else:
            sorted_filtered_models = {
                task: {arch: models for arch, models in archs.items()}
                for task, archs in filtered_models.items()
            }
        return sorted_filtered_models

    def _filter_task_types(self, task_types: List[str]):
        order = ["object detection", "instance segmentation", "pose estimation"]
        sorted_tt = [task for task in order if task in task_types]
        other_tasks = sorted(set(task_types) - set(order))
        sorted_tt.extend(other_tasks)
        return sorted_tt

    def set_models(self, models_list: List[Dict], sort_models: bool = False):
        self._models = models_list
        filtered_models = self._filter_and_sort_models(self._models, sort_models)
        self._table_data = filtered_models

        self._task_types = self._filter_task_types(list(filtered_models.keys()))
        self._arch_types = []
        # maintain correct order of arch types
        for task_type in self._task_types:
            for arch_type in filtered_models[task_type].keys():
                if arch_type not in self._arch_types:
                    self._arch_types.append(arch_type)

        self.__default_selected_arch_type = (
            self._arch_types[0] if len(self._arch_types) > 0 else None
        )
        self.__default_selected_task_type = (
            list(filtered_models[self.__default_selected_arch_type].keys())[0]
            if self.__default_selected_arch_type is not None
            else None
        )
        DataJson()[self.widget_id]["tableData"] = self._table_data
        DataJson().send_changes()

        StateJson()[self.widget_id]["selectedRow"] = 0
        StateJson()[self.widget_id]["taskTypes"] = self._task_types
        StateJson()[self.widget_id]["archTypes"] = self._arch_types
        StateJson()[self.widget_id]["selectedTaskType"] = self.__default_selected_task_type
        StateJson()[self.widget_id]["selectedArchType"] = self.__default_selected_arch_type
        StateJson().send_changes()

    def arch_type_changed(self, func):
        route_path = self.get_route_path(PretrainedModelsSelector.Routes.ARCH_TYPE_CHANGED)
        server = self._sly_app.get_server()
        self._arch_changes_handled = True

        @server.post(route_path)
        async def _arch_type_changed():
            res = self.get_selected_arch_type()
            func(res)

        return _arch_type_changed

    def task_type_changed(self, func):
        route_path = self.get_route_path(PretrainedModelsSelector.Routes.TASK_TYPE_CHANGED)
        server = self._sly_app.get_server()
        self._task_changes_handled = True

        @server.post(route_path)
        def _task_type_changed():
            res = self.get_selected_task_type()
            func(res)

        return _task_type_changed

    def model_changed(self, func):
        route_path = self.get_route_path(PretrainedModelsSelector.Routes.MODEL_CHANGED)
        server = self._sly_app.get_server()
        self._model_changes_handled = True

        @server.post(route_path)
        def _model_changed():
            res = self.get_selected_row()
            func(res)

        return _model_changed



from typing import Dict, List, Optional

try:
    from typing import Literal, Union
except ImportError:
    from typing_extensions import Literal

from supervisely.api.api import Api
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Checkbox, Empty, SelectProject, Widget, generate_id
from supervisely.app.widgets.select_sly_utils import _get_int_or_env
from supervisely.project.project_type import ProjectType
from supervisely.sly_logger import logger


class SelectDataset(Widget):
    """Widget to select one or multiple datasets, optionally coupled to a project selector."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        default_id: Union[int, List] = None,
        project_id: int = None,
        multiselect: bool = False,
        compact: bool = False,
        show_label: bool = True,
        size: Literal["large", "small", "mini"] = None,
        disabled: Optional[bool] = False,
        widget_id: str = None,
        select_all_datasets: bool = False,
        allowed_project_types: List[ProjectType] = [],
    ):
        """
        :param default_id: Initial dataset ID or list (if multiselect).
        :type default_id: Union[int, List], optional
        :param project_id: Project to list datasets from.
        :type project_id: int, optional
        :param multiselect: If True, allow multiple selection.
        :type multiselect: bool
        :param compact: If True, compact layout.
        :type compact: bool
        :param show_label: If True, show label.
        :type show_label: bool
        :param size: Size: "large", "small", or "mini".
        :type size: Literal["large", "small", "mini"], optional
        :param disabled: If True, widget is disabled.
        :type disabled: bool, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param select_all_datasets: If True (with multiselect), initial select-all.
        :type select_all_datasets: bool
        :param allowed_project_types: Filter by project types.
        :type allowed_project_types: List[ProjectType]
        """
        self._api = Api()
        self._default_id = default_id
        self._project_id = project_id
        self._multiselect = multiselect
        self._compact = compact
        self._show_label = show_label
        self._size = size
        self._team_selector = None
        self._all_datasets_checkbox = Empty()
        self._project_selector = Empty()
        self._project_types = allowed_project_types
        self._changes_handled = False
        self._cb_called = False
        self._disabled = disabled

        if self._multiselect is False:
            if isinstance(self._default_id, list):
                raise ValueError(
                    "Multiselect is disabled. To set 'default_id' use integers insted of a list of integers or switch multiselect to 'True'"
                )
        else:
            self._all_datasets_checkbox = Checkbox(
                "Select all datasets", checked=select_all_datasets, widget_id=generate_id()
            )

        self._default_id = _get_int_or_env(self._default_id, "modal.state.slyDatasetId")
        if self._default_id is not None:
            self._update_project_id()
        self._project_id = _get_int_or_env(self._project_id, "modal.state.slyProjectId")

        # NOW PROJECT CAN BE SET LATER WITH SET_PROJECT_ID METHOD
        # if compact is True:
        #     if self._project_id is None:
        #         raise ValueError(
        #             '"project_id" have to be passed as argument or "compact" has to be False'
        #         )
        # else:
        self._show_label = True
        self._project_selector = SelectProject(
            default_id=self._project_id,
            show_label=True,
            size=self._size,
            allowed_types=allowed_project_types,
            widget_id=generate_id(),
        )
        if self._disabled is True:
            self._project_selector.disable()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        res = {}
        res["disabled"] = self._disabled
        if self._compact is True:
            res["projectId"] = self._project_id
        res["options"] = {
            "showLabel": self._show_label,
            "compact": self._compact,
            "filterable": True,
            "valueProperty": "id",
            "multiple": self._multiselect,
            "flat": True,
            "availableProjectTypes": [ptype.value for ptype in self._project_types],
        }
        if self._size is not None:
            res["options"]["size"] = self._size
        return res

    def get_json_state(self) -> Dict:
        res = {"datasets": [self._default_id]}
        if self._compact is False:
            res["projectId"] = self._project_id
        return res

    def get_selected_project_id(self):
        if self._compact is True:
            raise ValueError(
                '"project_id" have to be passed as argument or "compact" has to be False'
            )
        return self._project_selector.get_selected_id()

    def get_selected_id(self):
        if self._multiselect is True:
            raise ValueError(
                "Multiselect is enabled. Use another method 'get_selected_ids' instead of 'get_selected_id'"
            )
        return StateJson()[self.widget_id]["datasets"]

    def get_selected_ids(self):
        if self._multiselect is False:
            raise ValueError(
                "Multiselect is disabled. Use another method 'get_selected_id' instead of 'get_selected_ids'"
            )
        if self._all_datasets_checkbox.is_checked():
            if self._compact is True:
                project_id = self._project_id
            else:
                project_id = self._project_selector.get_selected_id()
            if project_id is None:
                return [None]
            datasets = self._api.dataset.get_list(project_id)
            ids = [ds.id for ds in datasets]
            return ids
        else:
            return StateJson()[self.widget_id]["datasets"]

    def set_project_id(self, id: int):
        self._project_id = id
        if self._compact is True:
            DataJson()[self.widget_id]["projectId"] = self._project_id
            DataJson().send_changes()
        else:
            StateJson()[self.widget_id]["projectId"] = self._project_id
            StateJson().send_changes()
        self._project_selector.set_project_id(self._project_id)

    def set_select_all_datasets(self, is_checked: bool):
        if self._multiselect is False:
            raise ValueError(
                "Multiselect is disabled. Use another method 'set_dataset_id' instead of 'set_select_all_datasets'"
            )
        if is_checked:
            self._all_datasets_checkbox.check()
        else:
            self._all_datasets_checkbox.uncheck()
            StateJson()[self.widget_id]["datasets"] = []
            StateJson().send_changes()

    def set_dataset_id(self, id: int):
        if self._multiselect is True:
            raise ValueError(
                "Multiselect is enabled. Use another method 'set_dataset_ids' instead of 'set_dataset_id'"
            )
        elif isinstance(id, list):
            raise ValueError(
                "Enable multiselect and use another method 'set_dataset_ids' instead of 'set_dataset_id' to set list of ids"
            )
        self._default_id = id
        self._update_project_id()
        self.set_project_id(self._project_id)
        self._project_selector.set_project_id(self._project_id)
        StateJson()[self.widget_id]["datasets"] = self._default_id
        StateJson().send_changes()

    def set_dataset_ids(self, ids: Union[List[int], int]):
        if self._multiselect is False:
            raise ValueError(
                "Multiselect is disabled. Use another method 'set_dataset_id' instead of 'set_dataset_ids'"
            )
        if isinstance(ids, int):
            ids = [ids]
        self._default_id = ids
        self._update_project_id()
        self.set_project_id(self._project_id)
        self._project_selector.set_project_id(self._project_id)
        StateJson()[self.widget_id]["datasets"] = self._default_id
        StateJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(SelectDataset.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        def _process():
            if self._multiselect is True:
                value = self.get_selected_ids()
                if value == "":
                    value = None
                func(value)
            else:
                value = self.get_selected_id()
                if value == "":
                    value = None
                func(value)

        if self._multiselect is True:

            @self._all_datasets_checkbox.value_changed
            def _select_all_datasets(is_checked):
                if is_checked is False:
                    StateJson()[self.widget_id]["datasets"] = []
                    StateJson().send_changes()
                _process()

        if self._compact is False:

            @self._project_selector.value_changed
            def _update_datasets(project_id):
                if self._multiselect is True:
                    if project_id is not None and self._all_datasets_checkbox.is_checked():
                        datasets = self._api.dataset.get_list(project_id)
                        ids = [ds.id for ds in datasets]
                    else:
                        ids = []
                else:
                    ids = None
                StateJson()[self.widget_id]["datasets"] = ids
                StateJson().send_changes()
                self._cb_called = True
                _process()

        @server.post(route_path)
        def _click():
            if self._cb_called is False or self._multiselect is False:
                _process()
            elif self._multiselect and self._all_datasets_checkbox.is_checked() is False:
                _process()
            self._cb_called = False

        return _click

    def disable(self):
        self._project_selector.disable()
        self._all_datasets_checkbox.disable()
        self._disabled = True
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._all_datasets_checkbox.enable()
        self._project_selector.enable()
        self._disabled = False
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    @property
    def is_disabled(self) -> bool:
        return self._disabled

    @is_disabled.setter
    def is_disabled(self, value: int):
        self._disabled = value
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def _update_project_id(self):
        if isinstance(self._default_id, list) and len(self._default_id) != 0:
            info = self._api.dataset.get_info_by_id(self._default_id[0], raise_error=True)
        elif isinstance(self._default_id, int):
            info = self._api.dataset.get_info_by_id(self._default_id, raise_error=True)
        else:
            return
        self._project_id = info.project_id



# coding: utf-8
"""Create and manage labeling queues in Supervisely."""

# docs
from __future__ import annotations

from typing import Dict, List, Literal, NamedTuple, Optional, Union

from supervisely.api.constants import PLAYBACK_RATE_POSSIBLE_VALUES
from supervisely.api.labeling_job_api import LabelingJobApi
from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    ModuleWithStatus,
    RemoveableBulkModuleApi,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_type import ProjectType


class LabelingQueueInfo(NamedTuple):
    """NamedTuple describing a labeling queue returned by the API."""

    id: int
    name: str
    team_id: int
    project_id: int
    dataset_id: int
    created_by_id: int
    labelers: list
    reviewers: list
    created_at: str
    finished_at: str
    status: str
    jobs: list
    entities_count: int
    accepted_count: int
    annotated_count: int
    in_progress_count: int
    pending_count: int
    meta: dict
    collection_id: Optional[int] = None


class LabelingQueueApi(RemoveableBulkModuleApi, ModuleWithStatus):
    """
    API for working with labeling queues. :class:`~supervisely.api.labeling_queue_api.LabelingQueueApi` object is immutable.

    :Usage Example:

        .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            queue = api.labeling_queues.get_info_by_id(2) # api usage example
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple LabelingQueueInfo information about Labeling Queue.

        :Usage Example:

            .. code-block:: python

                LabelingQueueInfo(
                    id=2,
                    name='Annotation Queue (#1)',
                    team_id=4,
                    project_id=58,
                    dataset_id=54,
                    created_by_id=4,
                    labelers=[4],
                    reviewers=[4],
                    created_at='2020-04-08T15:10:12.618Z',
                    finished_at='2020-04-08T15:13:39.788Z',
                    status='completed',
                    jobs=[283, 282, 281],
                    entities_count=3,
                    accepted_count=2,
                    annotated_count=3,
                    in_progress_count=2,
                    pending_count=1,
                    meta={},
                    collection_id=None,
                )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.TEAM_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.CREATED_BY_ID,
            ApiField.LABELERS,
            ApiField.REVIEWERS,
            ApiField.CREATED_AT,
            ApiField.FINISHED_AT,
            ApiField.STATUS,
            ApiField.JOBS,
            ApiField.ENTITIES_COUNT,
            ApiField.ACCEPTED_COUNT,
            ApiField.ANNOTATED_COUNT,
            ApiField.IN_PROGRESS_COUNT,
            ApiField.PENDING_COUNT,
            ApiField.META,
            ApiField.COLLECTION_ID,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **LabelingQueueInfo**.
        """
        return "LabelingQueueInfo"

    def _convert_json_info(self, info: Dict, skip_missing: Optional[bool] = True):
        """ """

        def _get_value(dict, field_name, skip_missing):
            if skip_missing is True:
                return dict.get(field_name, None)
            else:
                return dict[field_name]

        def _get_ids(data, skip_missing):
            if skip_missing is True:
                return [job.get(ApiField.ID, None) for job in data]
            else:
                return [job[ApiField.ID] for job in data]

        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if type(field_name) is str:
                    if field_name in [ApiField.JOBS, ApiField.LABELERS, ApiField.REVIEWERS]:
                        field_values.append(_get_ids(info[field_name], skip_missing))
                        continue
                    else:
                        field_values.append(_get_value(info, field_name, skip_missing))
                elif type(field_name) is tuple:
                    value = None
                    for sub_name in field_name[0]:
                        if value is None:
                            value = _get_value(info, sub_name, skip_missing)
                        else:
                            value = _get_value(value, sub_name, skip_missing)
                    field_values.append(value)
                else:
                    raise RuntimeError("Can not parse field {!r}".format(field_name))
            res = self.InfoType(*field_values)
            return LabelingQueueInfo(**res._asdict())

    def _check_membership(self, ids: List[int], team_id: int) -> None:
        """Check if user is a member of the team in which the Labeling Queue is created."""
        for user_id in ids:
            memberships = self._api.user.get_teams(user_id)
            team_ids = [team.id for team in memberships]
            if team_id not in team_ids:
                raise RuntimeError(
                    f"User with id {user_id} is not a member of the team with id {team_id}"
                )

    def create(
        self,
        name: str,
        user_ids: List[int],
        reviewer_ids: List[int],
        dataset_id: Optional[int] = None,
        collection_id: Optional[int] = None,
        readme: Optional[str] = None,
        classes_to_label: Optional[List[str]] = None,
        objects_limit_per_image: Optional[int] = None,
        tags_to_label: Optional[List[str]] = None,
        tags_limit_per_image: Optional[int] = None,
        include_images_with_tags: Optional[List[str]] = None,
        exclude_images_with_tags: Optional[List[str]] = None,
        images_range: Optional[List[int, int]] = None,
        reviewer_id: Optional[int] = None,
        images_ids: Optional[List[int]] = None,
        dynamic_classes: Optional[bool] = False,
        dynamic_tags: Optional[bool] = False,
        disable_confirm: Optional[bool] = None,
        disable_submit: Optional[bool] = None,
        toolbox_settings: Optional[Dict] = None,
        hide_figure_author: Optional[bool] = False,
        allow_review_own_annotations: Optional[bool] = False,
        skip_complete_job_on_empty: Optional[bool] = False,
        enable_quality_check: Optional[bool] = None,
        quality_check_user_ids: Optional[List[int]] = None,
        guide_id: Optional[int] = None,
        description: Optional[str] = None,
    ) -> int:
        """
        Creates Labeling Queue and assigns given Users to it.

        :param name: Labeling Queue name in Supervisely.
        :type name: str
        :param user_ids: User IDs in Supervisely to assign Users as labelers to Labeling Queue.
        :type user_ids: List[int]
        :param reviewer_ids: User IDs in Supervisely to assign Users as reviewers to Labeling Queue.
        :type reviewer_ids: List[int]
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param collection_id: Entities Collection ID in Supervisely.
        :type collection_id: int, optional
        :param readme: Additional information about Labeling Queue.
        :type readme: str, optional
        :param classes_to_label: List of classes to label in :class:`~supervisely.project.project.Dataset`.
        :type classes_to_label: List[str], optional
        :param objects_limit_per_image: Limit the number of objects that the labeler can create on each image.
        :type objects_limit_per_image: int, optional
        :param tags_to_label: List of tags to label in :class:`~supervisely.project.project.Dataset`.
        :type tags_to_label: List[str], optional
        :param tags_limit_per_image: Limit the number of tags that the labeler can create on each image.
        :type tags_limit_per_image: int, optional
        :param include_images_with_tags: Include images with given tags for processing by labeler.
        :type include_images_with_tags: List[str], optional
        :param exclude_images_with_tags: Exclude images with given tags for processing by labeler.
        :type exclude_images_with_tags: List[str], optional
        :param images_range: Limit number of images to be labeled for each labeler.
        :type images_range: List[int, int], optional
        :param reviewer_id: User ID in Supervisely to assign User as Reviewer to Labeling Queue.
        :type reviewer_id: int, optional
        :param images_ids: List of images ids to label in dataset
        :type images_ids: List[int], optional
        :param dynamic_classes: If True, classes created after creating the queue will be available for annotators
        :type dynamic_classes: bool, optional
        :param dynamic_tags: If True, tags created after creating the queue will be available for annotators
        :type dynamic_tags: bool, optional
        :param disable_confirm: If True, the Confirm button will be disabled in the labeling tool. It will remain disabled until the next API call sets the parameter to False, re-enabling the button.
        :type disable_confirm: bool, optional
        :param disable_submit: If True, the Submit button will be disabled in the labeling tool. It will remain disabled until the next API call sets the parameter to False, re-enabling the button.
        :type disable_submit: bool, optional
        :param toolbox_settings: Settings for the labeling tool. Only video projects are supported.
        :type toolbox_settings: Dict, optional
        :param hide_figure_author: If True, hides the author of the figure in the labeling tool.
        :type hide_figure_author: bool, optional
        :param allow_review_own_annotations: If True, allows labelers to review their own annotations.
        :type allow_review_own_annotations: bool, optional
        :param skip_complete_job_on_empty: If True, skips completing the Labeling Queue if there are no images to label.
        :type skip_complete_job_on_empty: bool, optional
        :param enable_quality_check: If True, adds an intermediate step between "review" and completing the Labeling Queue.
        :type enable_quality_check: bool, optional
        :param quality_check_user_ids: List of User IDs in Supervisely to assign Users as Quality Checkers to Labeling Queue.
        :type quality_check_user_ids: List[int], optional
        :param guide_id: Guide ID in Supervisely to assign a guide to the Labeling Queue.
        :type guide_id: int, optional
        :param description: Description of Labeling Queue.
        :type description: str, optional
        :returns: Labeling Queue ID in Supervisely.
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                user_name = 'alex'
                dataset_id = 602
                new_labeling_queue_id = api.labeling_queue.create(
                    user_name,
                    dataset_id,
                    user_ids=[111, 222],
                    readme='Readmy text',
                    description='Work for labelers',
                    objects_limit_per_image=5,
                    tags_limit_per_image=3
                )
                print(new_labeling_queue_id)

                # >>> 2

                # Create video labeling job with toolbox settings

                user_id = 4
                dataset_id = 277
                video_id = 24897
                toolbox_settings = {"playbackRate": 32, "skipFramesSize": 15, "showVideoTime": True}

                new_labeling_queue_id = api.labeling_queue.create(
                    name="Labeling Queue name",
                    dataset_id=dataset_id,
                    user_ids=[user_id],
                    readme="Labeling Queue readme",
                    description="Some description",
                    classes_to_label=["car", "animal"],
                    tags_to_label=["animal_age_group"],
                    images_ids=[video_id],
                    toolbox_settings=toolbox_settings,
                )
                print(new_labeling_queue_id)

                # >>> 3
        """
        if dataset_id is None and collection_id is None:
            raise RuntimeError("Either dataset_id or collection_id must be provided")
        if classes_to_label is None:
            classes_to_label = []
        if tags_to_label is None:
            tags_to_label = []

        filter_images_by_tags = []
        if include_images_with_tags is not None:
            for tag_name in include_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": True})

        if exclude_images_with_tags is not None:
            for tag_name in exclude_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": False})

        if objects_limit_per_image is None:
            objects_limit_per_image = 0

        if tags_limit_per_image is None:
            tags_limit_per_image = 0

        meta = {
            "classes": classes_to_label,
            "projectTags": tags_to_label,
            "imageTags": filter_images_by_tags,
            "imageFiguresLimit": objects_limit_per_image,
            "imageTagsLimit": tags_limit_per_image,
            "dynamicClasses": dynamic_classes,
            "dynamicTags": dynamic_tags,
            "hideFigureAuthor": hide_figure_author,
        }
        if images_ids is not None:
            meta["entityIds"] = images_ids

        if guide_id is not None:
            try:
                guide_id = int(guide_id)
            except Exception as e:
                raise ValueError(
                    f"guide_id must be an integer, got {type(guide_id)} with value '{guide_id}'"
                ) from None
            meta["guide"] = guide_id

        if toolbox_settings is not None:
            if dataset_id is not None:
                dataset_info = self._api.dataset.get_info_by_id(dataset_id)
                project_id = dataset_info.project_id
            else:
                collection_info = self._api.entities_collection.get_info_by_id(collection_id)
                project_id = collection_info.project_id
            project_info = self._api.project.get_info_by_id(project_id)
            project_type = project_info.type
            if project_type == ProjectType.VIDEOS.value:
                playback_rate = toolbox_settings.get("playbackRate", None)
                if playback_rate is not None:
                    if playback_rate not in PLAYBACK_RATE_POSSIBLE_VALUES:
                        raise ValueError(
                            f"'playbackRate' must be one of: '{','.join(PLAYBACK_RATE_POSSIBLE_VALUES)}'"
                        )
                meta["toolboxSettings"] = toolbox_settings

        if disable_confirm is not None:
            meta.update({"disableConfirm": disable_confirm})
        if disable_submit is not None:
            meta.update({"disableSubmit": disable_submit})
        if enable_quality_check is not None:
            if quality_check_user_ids is None:
                raise RuntimeError(
                    "quality_check_user_ids must be provided if enable_quality_check is True"
                )
            meta.update({"enableIntermediateReview": enable_quality_check})

        queue_meta = {}
        if allow_review_own_annotations is True:
            queue_meta["reviewOwnAnnotationsAvailable"] = True

        if skip_complete_job_on_empty is True:
            queue_meta["skipCompleteAnnotationJobOnEmpty"] = True

        data = {
            ApiField.NAME: name,
            ApiField.USER_IDS: user_ids,
            ApiField.REVIEWER_IDS: reviewer_ids,
            ApiField.META: meta,
        }
        if dataset_id is not None:
            data[ApiField.DATASET_ID] = dataset_id
        if collection_id is not None:
            data[ApiField.COLLECTION_ID] = collection_id
        if quality_check_user_ids is not None:
            if enable_quality_check is not True:
                raise RuntimeError(
                    "quality_check_user_ids can be set only if enable_quality_check is True"
                )
            data[ApiField.QUALITY_CHECK_USER_IDS] = quality_check_user_ids

        if len(queue_meta) > 0:
            data[ApiField.QUEUE_META] = queue_meta

        if readme is not None:
            data[ApiField.README] = str(readme)

        if description is not None:
            data[ApiField.DESCRIPTION] = str(description)

        if images_range is not None and images_range != (None, None):
            if len(images_range) != 2:
                raise RuntimeError("images_range has to contain 2 elements (start, end)")
            images_range = {"start": images_range[0], "end": images_range[1]}
            data[ApiField.META]["range"] = images_range

        if reviewer_id is not None:
            data[ApiField.REVIEWER_ID] = reviewer_id

        response = self._api.post("labeling-queues.add", data)
        return response.json()["id"]  # {"success": true}

    def get_list(
        self,
        team_id: int,
        dataset_id: Optional[int] = None,
        project_id: Optional[int] = None,
        ids: Optional[List[int]] = None,
        names: Optional[List[str]] = None,
        show_disabled: Optional[bool] = False,
        collection_id: Optional[int] = None,
    ) -> List[LabelingQueueInfo]:
        """
        Get list of information about Labeling Queues in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int, optional
        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :param ids: List of Labeling Queue IDs in Supervisely.
        :type ids: List[int], optional
        :param names: List of Labeling Queue names in Supervisely.
        :type names: List[str], optional
        :param show_disabled: Show disabled Labeling Queues.
        :type show_disabled: bool, optional
        :param collection_id: Entities Collection ID in Supervisely.
        :type collection_id: int, optional
        :returns: List of information about Labeling Queues.
        :rtype: :class:`List[LabelingQueueInfo]`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                label_jobs = api.labeling_queue.get_list(4)
        """

        filters = []
        if project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": project_id})
        if dataset_id is not None:
            filters.append({"field": ApiField.DATASET_ID, "operator": "=", "value": dataset_id})
        if collection_id is not None:
            filters.append(
                {"field": ApiField.COLLECTION_ID, "operator": "=", "value": collection_id}
            )
        if names is not None:
            filters.append({"field": ApiField.NAME, "operator": "in", "value": names})
        if ids is not None:
            filters.append({"field": ApiField.ID, "operator": "in", "value": ids})
        return self.get_list_all_pages(
            "labeling-queues.list",
            {ApiField.TEAM_ID: team_id, "showDisabled": show_disabled, ApiField.FILTER: filters},
        )

    def get_info_by_id(self, id: int) -> LabelingQueueInfo:
        """
        Get Labeling Queue information by ID.

        :param id: Labeling Queue ID in Supervisely.
        :type id: int
        :returns: Information about Labeling Queue.
        :rtype: :class:`~supervisely.api.labeling_queue_api.LabelingQueueInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                label_job_info = api.labeling_queue.get_info_by_id(2)
                print(label_job_info)
                # Output: [
                #     2,
                #     "Annotation Queue (#1) (#1) (dataset_01)",
                #     "",
                #     "",
                #     4,
                #     8,
                #     "First Workspace",
                #     58,
                #     "tutorial_project",
                #     54,
                #     "dataset_01",
                #     4,
                #     "anna",
                #     4,
                #     "anna",
                #     4,
                #     "anna",
                #     "2020-04-08T15:10:12.618Z",
                #     "2020-04-08T15:10:19.833Z",
                #     "2020-04-08T15:13:39.788Z",
                #     "completed",
                #     false,
                #     3,
                #     0,
                #     1,
                #     2,
                #     2,
                #     [],
                #     [],
                #     [
                #         1,
                #         5
                #     ],
                #     null,
                #     null,
                #     [],
                #     [],
                #     [],
                #     [
                #         {
                #             "reviewStatus": "rejected",
                #             "id": 283,
                #             "name": "image_03"
                #         },
                #         {
                #             "reviewStatus": "accepted",
                #             "id": 282,
                #             "name": "image_02"
                #         },
                #         {
                #             "reviewStatus": "accepted",
                #             "id": 281,
                #             "name": "image_01"
                #         }
                #     ]
                # ]
        """
        return self._get_info_by_id(id, "labeling-queues.info")

    def get_status(self, id: int) -> LabelingJobApi.Status:
        """
        Get status of Labeling Job with given ID.

        :param id: Labeling job ID in Supervisely.
        :type id: int
        :returns: Labeling Job Status
        :rtype: :class:`~supervisely.api.labeling_job_api.LabelingJobApi.Status`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                job_status = api.labeling_queue.get_status(4)
                print(job_status) # pending
        """
        status_str = self.get_info_by_id(id).status
        return LabelingJobApi.Status(status_str)

    def set_status(self, id: int, status: str) -> None:
        """
        Sets Labeling Queue status.

        :param id: Labeling Queue ID in Supervisely.
        :type id: int
        :param status: New Labeling Queue status
        :type status: str
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()
                api.labeling_queue.set_status(id=9, status="completed")
        """
        self._api.post("labeling-queues.set-status", {ApiField.ID: id, ApiField.STATUS: status})

    def get_project_meta(self, id: int) -> ProjectMeta:
        """
        Returns project meta with classes and tags used in the labeling queue with given id.

        :param id: Labeling Queue ID in Supervisely.
        :type id: int
        :returns: Project meta of the labeling queue with given id.
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        queue_info = self.get_info_by_id(id)
        project_meta_json = self._api.project.get_meta(queue_info.project_id, with_settings=True)
        project_meta = ProjectMeta.from_json(project_meta_json)

        jobs = [self._api.labeling_job.get_info_by_id(job_id) for job_id in queue_info.jobs]
        job_classes = set()
        job_tags = set()
        for job in jobs:
            job_classes.update(job.classes_to_label)
            job_tags.update(job.tags_to_label)

        filtered_classes = [
            obj_cls for obj_cls in project_meta.obj_classes if obj_cls.name in job_classes
        ]
        filtered_tags = [
            tag_meta for tag_meta in project_meta.tag_metas if tag_meta.name in job_tags
        ]
        queue_meta = ProjectMeta(obj_classes=filtered_classes, tag_metas=filtered_tags)
        return queue_meta

    def get_entities_all_pages(
        self,
        id: int,
        collection_id: Optional[int] = None,
        per_page: int = 25,
        sort: str = "name",
        sort_order: str = "asc",
        status: Optional[Union[List, Literal["none", "done", "accepted", "null"]]] = None,
        limit: int = None,
        filter_by: List[Dict] = None,
    ) -> Dict[str, Union[List[Dict], int]]:
        """
        Get list of all or limited quantity entities from the Supervisely server.

        :param id: Labeling Queue ID in Supervisely.
        :type id: int
        :param collection_id: Collection ID in Supervisely.
        :type collection_id: int, optional
        :param per_page: Number of entities per page.
        :type per_page: int, optional
        :param sort: Sorting field.
        :type sort: str, optional
        :param sort_order: Sorting order.
        :type sort_order: str, optional
        :param status: Status of entities to filter. Possible values:

            - ``"null"`` - pending (in queue)
            - ``"none"`` - annotating (not in queue)
            - ``"done"`` - on review
            - ``"accepted"`` - accepted
        :type status: str or List[str], optional
        :param limit: Limit the number of entities to return. If limit is None, all entities will be returned.
        :type limit: int, optional
        :param filter_by: Filter for entities. Each element is a dict with keys ``field``, ``operator``, ``value``.
            Example: ``[{"field": "name", "operator": "in", "value": ["image_01", "image_02"]}]``.
        :type filter_by: List[Dict], optional
        """

        method = "labeling-queues.stats.entities"
        data = {
            ApiField.ID: id,
            ApiField.PAGE: 1,
            ApiField.PER_PAGE: per_page,
            ApiField.SORT: sort,
            ApiField.SORT_ORDER: sort_order,
        }
        if collection_id is not None:
            data[ApiField.FILTERS] = [
                {
                    "type": "entities_collection",
                    "data": {ApiField.COLLECTION_ID: collection_id, ApiField.INCLUDE: True},
                }
            ]
        if filter_by is not None:
            data[ApiField.FILTER] = filter_by
        if status is not None:
            if type(status) is str:
                status = [status]
            status = [None if s == "null" else s.lower() for s in status]
            data["entityStatus"] = status

        first_response = self._api.post(method, data).json()
        total = first_response["total"]
        pages_count = int(total / per_page) + 1 if total % per_page != 0 else int(total / per_page)
        if pages_count in [0, 1]:
            return first_response

        limit_exceeded = False
        results = first_response["images"]
        if limit is not None and len(results) > limit:
            limit_exceeded = True

        if (pages_count == 1 and len(results) == total) or limit_exceeded is True:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, "page": page_idx, "per_page": per_page})
                temp_items = temp_resp.json()["images"]
                results.extend(temp_items)
                if limit is not None and len(results) > limit:
                    limit_exceeded = True
                    break

            if len(results) != total and limit is None:
                raise RuntimeError(
                    "Method {!r}: error during pagination, some items are missed".format(method)
                )

        if limit is not None:
            results = results[:limit]
        return {"images": results, "total": total}

    def get_entities_count_by_status(
        self,
        id: int,
        status: Optional[Union[List, Literal["none", "done", "accepted", "null"]]] = None,
        filter_by: List[Dict] = None,
    ) -> int:
        """
        Get count of entities in the given Labeling Queue with given status.

        :param id: Labeling Queue ID in Supervisely.
        :type id: int
        :param status: Status of entities to filter. Possible values:

            - ``"null"`` - pending (in queue)
            - ``"none"`` - annotating (not in queue)
            - ``"done"`` - on review
            - ``"accepted"`` - accepted
        :type status: str or List[str], optional
        :param filter_by: Filter for entities. Each element is a dict with keys ``field``, ``operator``, ``value``.
            Example: ``[{"field": "name", "operator": "in", "value": ["image_01", "image_02"]}]``.
        :type filter_by: List[Dict], optional
        :returns: Count of entities in the Labeling Queue with given status.
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                entities_count = api.labeling_queue.get_entities_count_by_status(4, status="none")
                print(entities_count)
                # Output: 3
        """
        return self.get_entities_all_pages(id, status=status, limit=1, filter_by=filter_by).get(
            "total", 0
        )



# coding: utf-8
"""Work with video figures via the Supervisely API."""

# docs
from __future__ import annotations

from typing import Dict, List, Optional

from supervisely.api.entity_annotation.figure_api import FigureApi, FigureInfo
from supervisely.api.module_api import ApiField
from supervisely.geometry.geometry import Geometry
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.annotation.label import LabelingStatus

class VideoFigureApi(FigureApi):
    """
    API for working with :class:`~supervisely.video_annotation.video_figure.VideoFigure`.
    :class:`~supervisely.api.video.video_figure_api.VideoFigureApi` object is immutable.
    """

    def create(
        self,
        video_id: int,
        object_id: int,
        frame_index: int,
        geometry_json: dict,
        geometry_type: str,
        track_id: Optional[int] = None,
        meta: Optional[dict] = None,
        status: Optional[LabelingStatus] = None,
    ) -> int:
        """
        Create new VideoFigure for given frame in given video ID.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param object_id: ID of the object to which the VideoFigure belongs.
        :type object_id: int
        :param frame_index: Number of the frame to add :class:`~supervisely.video_annotation.video_figure.VideoFigure`.
        :type frame_index: int
        :param geometry_json: Parameters of geometry for :class:`~supervisely.video_annotation.video_figure.VideoFigure`.
        :type geometry_json: dict
        :param geometry_type: Type of VideoFigure geometry.
        :type geometry_type: str
        :param track_id: int, optional.
        :type track_id: int, optional
        :param meta: Meta data for :class:`~supervisely.video_annotation.video_figure.VideoFigure`.
        :type meta: dict, optional
        :param status: Labeling status. Specifies if the VideoFigure was created by NN model, manually or created by NN and then manually corrected.
        :type status: :class:`~supervisely.annotation.label.LabelingStatus`, optional
        :returns: New figure ID
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                video_id = 198703211
                object_id = 152118
                frame_idx = 0
                geometry_json = {'points': {'exterior': [[500, 500], [1555, 1500]], 'interior': []}}
                geometry_type = 'rectangle'

                figure_id = api.video.figure.create(video_id, object_id, frame_idx, geometry_json, geometry_type) # 643182610
        """
        if meta is None:
            meta = {}
        meta = {**(meta or {}), ApiField.FRAME: frame_index}

        return super().create(
            video_id,
            object_id,
            meta,
            geometry_json,
            geometry_type,
            track_id,
            status=status,
        )

    def append_bulk(self, video_id: int, figures: List[VideoFigure], key_id_map: KeyIdMap) -> None:
        """
        Add VideoFigures to given Video by ID.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param figures: List of VideoFigures to append.
        :type figures: List[:class:`~supervisely.video_annotation.video_figure.VideoFigure`]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 124976
                meta_json = api.project.get_meta(project_id)
                meta = sly.ProjectMeta.from_json(meta_json)
                key_id_map = KeyIdMap()

                video_id = 198703212
                ann_info = api.video.annotation.download(video_id)
                ann = sly.VideoAnnotation.from_json(ann_info, meta, key_id_map)
                figures = ann.figures[:5]
                api.video.figure.append_bulk(video_id, figures, key_id_map)
        """

        keys = []
        figures_json = []
        for figure in figures:
            keys.append(figure.key())
            figures_json.append(figure.to_json(key_id_map, save_meta=True))

        self._append_bulk(video_id, figures_json, keys, key_id_map)

    def update(self, figure_id: int, geometry: Geometry, status: Optional[LabelingStatus] = None) -> None:
        """Updates figure feometry with given ID in Supervisely with new Geometry object.

        :param figure_id: ID of the figure to update
        :type figure_id: int
        :param geometry: Supervisely Gepmetry object
        :type geometry: :class:`~supervisely.geometry.geometry.Geometry`
        :param status: Labeling status. Specifies if the VideoFigure was created by NN model, manually or created by NN and then manually corrected.
        :type status: :class:`~supervisely.annotation.label.LabelingStatus`, optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                new_geometry: sly.Rectangle(10, 10, 100, 100)
                figure_id = 121236918

                api.video.figure.update(figure_id, new_geometry)
        """
        payload = {
            ApiField.ID: figure_id,
            ApiField.GEOMETRY: geometry.to_json(),
        }

        if status is not None:
            nn_created,nn_updated = LabelingStatus.to_flags(status)
            payload[ApiField.NN_CREATED] = nn_created
            payload[ApiField.NN_UPDATED] = nn_updated

        self._api.post("figures.editInfo", payload)

    def download(
        self, dataset_id: int, video_ids: List[int] = None, skip_geometry: bool = False, **kwargs
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of video ID and list of FigureInfo for the given dataset ID. Can be filtered by video IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param video_ids: Specify the list of video IDs within the given dataset ID. If video_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type video_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool
        :returns: A dictionary where keys are video IDs and values are lists of figures.
        :rtype: Dict[int, List[:class:`~supervisely.api.entity_annotation.figure_api.FigureInfo`]]
        """
        if kwargs.get("image_ids", False) is not False:
            video_ids = kwargs["image_ids"]  # backward compatibility
        return super().download(dataset_id, video_ids, skip_geometry)



# coding: utf-8
from supervisely.annotation.renamer import Renamer
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection


class TagMetaMapper:
    """Strategy interface for mapping a source :class:`~supervisely.annotation.tag_meta.TagMeta` to a destination one."""

    def map(self, src: TagMeta) -> TagMeta:
        """
        """
        raise NotImplementedError()


class RenamingTagMetaMapper(TagMetaMapper):
    """Map tag metas by renaming their names using a :class:`~supervisely.annotation.renamer.Renamer`."""

    def __init__(self, dest_tag_meta_dict: TagMetaCollection, renamer: Renamer):
        """
        :param dest_tag_meta_dict: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection` object to map to.
        :type dest_tag_meta_dict: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param renamer: :class:`~supervisely.annotation.renamer.Renamer` object to use for renaming.
        :type renamer: :class:`~supervisely.annotation.renamer.Renamer`
        """
        self._dest_tag_meta_dict = dest_tag_meta_dict
        self._renamer = renamer

    def map(self, src: TagMeta) -> TagMeta:
        """
        The function map rename TagMeta in given collection
        :returns: TagMeta object
        :rtype: :class:`~supervisely.annotation.tag_meta.TagMeta`
        """
        dest_name = self._renamer.rename(src.name)
        return self._dest_tag_meta_dict.get(dest_name, None) if (dest_name is not None) else None


def make_renamed_tags(tags: TagCollection, tag_meta_mapper: TagMetaMapper, skip_missing=True) -> TagCollection:
    """
    The function make_renamed_tags rename tags names in given collection and return new collection
    :returns: TagCollection object
    :rtype: :class:`~supervisely.annotation.tag_collection.TagCollection`
    """
    renamed_tags = []
    for tag in tags:
        dest_tag_meta = tag_meta_mapper.map(tag.meta)
        if dest_tag_meta is not None:
            renamed_tags.append(tag.clone(meta=dest_tag_meta))
        elif not skip_missing:
            raise KeyError('Tag named {} could not be mapped to a destination name.'.format(tag.meta.name))
    return TagCollection(items=renamed_tags)



from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class RenormalizedErrorOverUnion(SemanticSegmVisMetric):
    """Renormalized boundary/extent/segment error-over-union charts for segmentation reports."""


    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "renormalized_error_over_union",
            "Renormalized Error over Union",
            text=self.vis_texts.markdown_renormalized_error_ou,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("intersection_error_over_union", self.get_figure())

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # Renormalized Error over Union figure
        labels = ["Boundary EoU", "Extent EoU", "Segment EoU"]
        values = [
            self.eval_result.mp.boundary_renormed_eou,
            self.eval_result.mp.extent_renormed_eou,
            self.eval_result.mp.segment_renormed_eou,
        ]
        fig.add_trace(
            go.Bar(
                x=labels,
                y=values,
                orientation="v",
                text=values,
                width=[0.5, 0.5, 0.5],
                textposition="outside",
                marker_color=["#FFE4B5", "#F7ADAA", "#dd3f3f"],
                hovertemplate="%{x}: %{y:.2f}<extra></extra>",
            )
        )
        fig.update_traces(hovertemplate="%{x}: %{y:.2f}<extra></extra>")
        fig.update_layout(width=600)

        return fig



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class Sidebar(Widget):
    """Collapsible sidebar layout: left panel (narrow) and main content area; supports open/close toggle."""

    def __init__(
        self,
        left_content: Widget,
        right_content: Widget,
        width_percent: int = 25,
        widget_id: str = None,
        height: str = "100vh",
        standalone: bool = True,
        clear_main_panel_paddings: bool = False,
        show_close: bool = True,
        show_open: bool = True,
        sidebar_left_padding: str = None,
    ):
        """:param left_content: Widget for sidebar (left panel).
        :type left_content: Widget
        :param right_content: Widget for main content area.
        :type right_content: Widget
        :param width_percent: Sidebar width as percent (0-100).
        :type width_percent: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param height: CSS height (e.g. "100vh").
        :type height: str
        :param standalone: If True, clear body padding.
        :type standalone: bool
        :param clear_main_panel_paddings: If True, no padding on main panel.
        :type clear_main_panel_paddings: bool
        :param show_close: If True, show close button.
        :type show_close: bool
        :param show_open: If True, show open button when collapsed.
        :type show_open: bool
        :param sidebar_left_padding: CSS padding for sidebar.
        :type sidebar_left_padding: str, optional
        """
        self._left_content = left_content
        self._right_content = right_content
        self._width_percent = width_percent
        self._options = {
            "sidebarWidth": self._width_percent,
            "height": height,
            "clearMainPanelPaddings": clear_main_panel_paddings,
            "showOpen": show_open,
            "showClose": show_close,
            "sidebarLeftPadding": sidebar_left_padding,
        }
        super().__init__(widget_id=widget_id, file_path=__file__)

        if standalone:
            StateJson()["app_body_padding"] = "0px"
        StateJson()["menuIndex"] = "1"

    def get_json_data(self):
        return {"options": self._options}

    def get_json_state(self):
        return {
            "sidebarCollapsed": False,
        }

    def collapse(self):
        StateJson()[self.widget_id]["sidebarCollapsed"] = True
        StateJson().send_changes()

    def expand(self):
        StateJson()[self.widget_id]["sidebarCollapsed"] = False
        StateJson().send_changes()



from __future__ import annotations

import pandas as pd

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class PRCurveByClass(DetectionVisMetric):
    """Per-class precision-recall curves for detection reports."""

    MARKDOWN = "pr_curve_by_class"
    CHART = "pr_curve_by_class"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_pr_by_class
        return MarkdownWidget(self.MARKDOWN, "PR Curve by Class", text)

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget(self.CHART, self._get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].data.legendgroup}`,",
        )
        return chart

    def _get_figure(self):  # -> go.Figure:
        import plotly.express as px  # pylint: disable=import-error

        df = pd.DataFrame(self.eval_result.mp.pr_curve(), columns=self.eval_result.mp.cat_names)

        fig = px.line(
            df,
            x=self.eval_result.mp.recThrs,
            y=df.columns,
            labels={"x": "Recall", "value": "Precision", "variable": "Category"},
            color_discrete_sequence=px.colors.qualitative.Prism,
            width=800,
            height=600,
        )

        fig.update_yaxes(range=[0, 1])
        fig.update_xaxes(range=[0, 1])
        return fig



from typing import Any
from supervisely.app.widgets import Widget


class Draggable(Widget):
    """Wrapper widget that makes its content draggable within the UI."""

    def __init__(self, content: Widget, key: Any = None, widget_id: str = None):
        """
        :param content: Widget to make draggable.
        :type content: :class:`~supervisely.app.widgets.widget.Widget`
        :param key: Optional key for drag state. Defaults to content.widget_id.
        :type key: Any, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._content = content
        self._key = key
        if key is None:
            self._key = content.widget_id
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Union

from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Text, Widget


class CheckboxField(Widget):
    """Checkbox with title and description."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        title: str,
        description: str,
        checked: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """
        :param title: Checkbox title.
        :type title: str
        :param description: Checkbox description.
        :type description: str
        :param checked: Initial checked state.
        :type checked: bool, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import CheckboxField
                cb = CheckboxField("Title", "Description", checked=False)
        """
        self._title = title
        self._description = description
        self._checked = checked
        self._changes_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        """Returns the data of the checkbox.
        Dictionary contains the following keys:
            - title: Current title data of the checkbox.
            - description: Current description data of the checkbox

        :returns: Dictionary with the data of the checkbox.
        :rtype: Dict[str, str]
        """
        return {"title": self._title, "description": self._description}

    def get_json_state(self) -> Dict[str, bool]:
        """Returns the state of the checkbox.
        Dictionary contains the following keys:
            - checked: Current state of the checkbox.

        :returns: Dictionary with the state of the checkbox.
        :rtype: Dict[str, bool]
        """
        return {"checked": self._checked}

    def is_checked(self) -> bool:
        """Returns the state of the checkbox. True if checked, False otherwise.

        :returns: Current state of the checkbox.
        :rtype: bool
        """
        return StateJson()[self.widget_id]["checked"]

    def set(self, title: str = None, description: str = None, checked: bool = None) -> None:
        """Set title, description and checkbox state to widget

        :param title: Title of the checkbox.
        :type title: Optional[str]
        :param description: Description of the checkbox.
        :type description: Optional[str]
        :param checked: Initial state of the checkbox.
        :type checked: Optional[bool]
        :param checked: New state of the checkbox.
        :type checked: bool
        """
        if checked is not None:
            self._checked = checked
            StateJson()[self.widget_id]["checked"] = self._checked
            StateJson().send_changes()
        if title is not None:
            self._title = title
            DataJson()[self.widget_id]["title"] = self._title
        if description is not None:
            self._description = description
            DataJson()[self.widget_id]["description"] = self._description
        if title is not None or description is not None:
            DataJson().send_changes()

    def check(self) -> None:
        """Sets the state of the checkbox to True."""
        self._checked = True
        StateJson()[self.widget_id]["checked"] = self._checked
        StateJson().send_changes()

    def uncheck(self) -> None:
        """Sets the state of the checkbox to False."""
        self._checked = False
        StateJson()[self.widget_id]["checked"] = self._checked
        StateJson().send_changes()

    def value_changed(self, func: Callable[[bool], Any]) -> Callable[[], None]:
        """Decorator that handles the event of changing the state of the checkbox.

        :param func: Function that handles the event of changing the state of the checkbox.
        :type func: Callable[[bool], Any]
        :returns: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(CheckboxField.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.is_checked()
            func(res)

        return _click



import asyncio
import copy
import math
import re
import sys
import weakref
from functools import partial
from typing import Any, Union

from tqdm import tqdm

from supervisely import is_development, is_production

# from supervisely import _original_tqdm as tqdm
from supervisely.app import DataJson
from supervisely.app.fastapi import run_sync
from supervisely.app.singleton import Singleton
from supervisely.app.widgets import Widget
from supervisely.sly_logger import EventType, logger


def extract_by_regexp(regexp, string):
    result = re.search(regexp, string)
    if result is not None:
        return result.group(0)
    else:
        return ""


UNITS = ["B", "KB", "MB", "GB", "TB"]


class _slyProgressBarIO:
    """File-like adapter used by tqdm to report progress into Supervisely `DataJson` and logs."""

    def __init__(self, widget_id, message=None, total=None, unit=None, unit_scale=None):
        """:param widget_id: Progress widget ID.
        :param message: Progress message.
        :param total: Total iterations/size.
        :param unit: Unit string (e.g. "it", "B").
        :param unit_scale: If True, scale units (B, KB, MB...).
        """
        self.widget_id = widget_id

        self.progress = {
            "percent": 0,
            "info": "",
            "message": message,
            "status": None,
            "n": -1,
        }

        self.prev_state = self.progress.copy()
        self.total = total

        self.unit_scale = unit_scale
        self.unit = unit
        if self.unit_scale and self.unit == "B":
            self.unit = self.get_unit()

        self._n = 0

        self._done = False

    def print_progress_to_supervisely_tasks_section(self):
        """
        Logs a message with level INFO on logger. Message contain type of progress, subtask message, currtnt and total number of iterations
        """

        if self._n == self.progress["n"]:
            return

        self.progress["n"] = self._n

        extra = {
            "event_type": EventType.PROGRESS,
            "subtask": self.progress.get("message", None),
            "current": self._n,
            "total": self.total,
        }

        if self.unit_scale and self.unit != "it":
            extra["current_label"] = f"{self.bytes_to_unit(self._n)} {self.unit}"
            extra["total_label"] = f"{self.bytes_to_unit(self.total)} {self.unit}"

        gettrace = getattr(sys, "gettrace", None)
        in_debug_mode = gettrace is not None and gettrace()

        if not in_debug_mode:
            logger.info("progress", extra=extra)

    def write(self, s):
        new_text = s.strip().replace("\r", "")
        if len(new_text) != 0:
            if self.total is not None:
                if self.total == 0:  # to avoid ZeroDivisionError
                    self.progress["percent"] = 100
                else:
                    self.progress["percent"] = int(self._n / self.total * 100)
                self.progress["info"] = extract_by_regexp(r"(\d+(?:\.\d+\w+)?)*\w*/.*\]", new_text)
            else:
                self.progress["percent"] = int(self._n)
                self.progress["info"] = extract_by_regexp(r"(\d+(?:\.\d+\w+)?)*.*\]", new_text)

    def flush(self, synchronize_changes=True):
        if self.prev_state != self.progress:
            if self.progress["percent"] != "" and self.progress["info"] != "":
                self.print_progress_to_supervisely_tasks_section()

                if self.progress["percent"] == 100 and self.total is not None:
                    self.progress["status"] = "success"

                for key, value in self.progress.items():
                    DataJson()[f"{self.widget_id}"][key] = value

                if synchronize_changes is True:
                    run_sync(DataJson().synchronize_changes())

                self.prev_state = copy.deepcopy(self.progress)

    def __del__(self):
        self.progress["status"] = "success"
        self.progress["percent"] = 100

        self.flush(synchronize_changes=False)
        self.print_progress_to_supervisely_tasks_section()

    def get_unit(self) -> str:
        """Returns the unit of the progress bar according to the total number of bytes.

        :returns: unit of the progress bar
        :rtype: str
        """
        total = self.total
        for unit in UNITS:
            if total < 1000:
                return unit
            total /= 1000

    def bytes_to_unit(self, bytes: int) -> Union[float, Any]:
        """Returns a human-readable string representation of bytes according to the
        self.unit.

        :param bytes: number of bytes
        :type bytes: int
        :returns: converted size from bytes to self.unit
        :rtype: Union[float, Any]
        """
        if not isinstance(bytes, int) or self.unit not in UNITS:
            return bytes

        for idx, unit in enumerate(UNITS):
            if self.unit == unit:
                return round(bytes / (1000**idx), 2)


class CustomTqdm(tqdm):
    """tqdm subclass that streams progress updates to a Supervisely progress-bar widget."""

    def __init__(self, widget_id, message, *args, **kwargs):
        """:param widget_id: Progress widget ID to sync with.
        :param message: Progress message.
        :param args: Passed to tqdm (e.g. iterable).
        :param kwargs: Passed to tqdm (e.g. total, unit, unit_scale).
        """
        extracted_total = copy.copy(
            tqdm(iterable=kwargs["iterable"], total=kwargs["total"], disable=True).total
        )
        unit_scale = kwargs.get("unit_scale")
        unit = kwargs.get("unit")

        self._iteration_value = 0
        self._iteration_number = 0
        self._iteration_locked = False
        self._total_monitor_size = 0

        self.unit_divisor = 1024

        # self.n = 0
        # self.reported_cnt = 0
        # self.report_every = max(1, math.ceil(self.extracted_total / 100))

        super().__init__(
            file=_slyProgressBarIO(widget_id, message, extracted_total, unit, unit_scale),
            *args,
            **kwargs,
        )

    def __setattr__(self, name, value):
        if name == "total":
            if hasattr(self, "fp"):
                self.fp.total = value
        if name == "message":
            if hasattr(self, "fp"):
                self.fp.progress["message"] = value
        return super().__setattr__(name, value)

    def refresh(self, *args, **kwargs):
        if self.fp is not None:
            self.fp._n = self.n
        super().refresh(*args, **kwargs)

    def close(self):
        self.refresh()
        super(CustomTqdm, self).close()
        if self.fp is not None:
            self.fp.__del__()

    def __del__(self):
        super(CustomTqdm, self).__del__()
        if self.fp is not None:
            self.fp.__del__()

    def _progress_monitor(self, monitor):
        if is_development() and self.n >= self.total:
            self.refresh()
            self.close()

        if monitor.bytes_read == 8192:
            self._total_monitor_size += monitor.len

        if self._total_monitor_size > self.total:
            self.total = self._total_monitor_size

        if not self._iteration_locked:
            # if is_development():
            super().update(self._iteration_value + monitor.bytes_read - self.n)
            # else:
            #     self.set_current_value(self._iteration_value + monitor.bytes_read, report=False)

        if monitor.bytes_read == monitor.len and not self._iteration_locked:
            self._iteration_value += monitor.len
            self._iteration_number += 1
            self._iteration_locked = True
            if is_development():
                self.refresh()

        if monitor.bytes_read < monitor.len:
            self._iteration_locked = False

    def get_partial(self):
        return partial(self._progress_monitor)


class SlyTqdm(Widget):
    """Widget wrapper around tqdm that renders progress in the Supervisely UI."""

    # @TODO: track all active sessions for one object and close them if new object inited
    def __init__(self, message: str = None, show_percents: bool = False, widget_id: str = None):
        """
        Wrapper for classic tqdm progress bar.

            Parameters
            ----------
            identifier  : int, required
                HTML element identifier
            message  : int, optional
                Text message which displayed in HTML


            desc, total, leave, ncols, ... :
                Like in tqdm

        """
        self.message = message
        self.show_percents = show_percents

        self._active_session = None
        self._sly_io = None
        self._hide_on_finish = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _close_active_session(self):
        if self._active_session is not None:
            try:
                self._active_session.__del__()
                self._active_session = None
            except ReferenceError:
                pass

    def __call__(
        self,
        iterable=None,
        message=None,
        desc=None,
        total=None,
        leave=None,
        ncols=None,
        mininterval=1.0,
        maxinterval=10.0,
        miniters=None,
        ascii=False,
        disable=False,
        unit="it",
        unit_scale=False,
        dynamic_ncols=False,
        smoothing=0.3,
        bar_format=None,
        initial=0,
        position=None,
        postfix=None,
        unit_divisor=1000,
        gui=False,
        **kwargs,
    ):
        return CustomTqdm(
            widget_id=self.widget_id,
            iterable=iterable,
            desc=desc,
            total=total,
            leave=leave,
            message=message,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            gui=gui,
            **kwargs,
        )

    def get_json_data(self):
        return {
            "percent": 0,
            "info": None,
            "message": self.message,
            "status": None,
            "show_percents": self.show_percents,
        }

    def get_json_state(self):
        return None

    def set_message(self, message):
        self.message = message
        DataJson()[self.widget_id]["message"] = message
        DataJson().send_changes()


class Progress(SlyTqdm):
    """Convenience progress widget that can automatically hide itself when finished."""

    def __init__(
        self,
        message: str = None,
        show_percents: bool = False,
        hide_on_finish=True,
        widget_id: str = None,
    ):
        """:param message: Progress message.
        :type message: str, optional
        :param show_percents: If True, show percentage.
        :type show_percents: bool
        :param hide_on_finish: If True, hide widget when done.
        :type hide_on_finish: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self.hide_on_finish = hide_on_finish
        super().__init__(message=message, show_percents=show_percents, widget_id=widget_id)

    pass
